from __future__ import annotations

import math

from scitbx.array_family import flex

from dials_refinement_helpers_ext import maha_dist_sq as maha_dist_sq_cpp
from dials_refinement_helpers_ext import mcd_consistency


def sample_covariance(a, b):
    """Calculate sample covariance of two vectors"""

    N = len(a)
    assert len(b) == N
    return flex.sum((a - flex.mean(a)) * (b - flex.mean(b))) / (N - 1)


def cov(*args):
    """Calculate covariance matrix between the arguments (should be flex.double
    arrays of equal length)"""

    lens = [len(e) for e in args]
    assert all(e == lens[0] for e in lens)

    ncols = len(args)
    cov = flex.double(flex.grid(ncols, ncols))
    for i in range(ncols):
        for j in range(i, ncols):
            cov[i, j] = sample_covariance(args[i], args[j])

    cov.matrix_copy_upper_to_lower_triangle_in_place()
    return cov


def maha_dist_sq(cols, center, cov):
    """Calculate squared Mahalanobis distance of all observations (rows in the
    vectors contained in the list cols) from the center vector with respect to
    the covariance matrix cov"""

    n = len(cols[0])
    p = len(cols)
    assert len(center) == p

    # observation matrix
    obs = flex.double(flex.grid(n, p))
    for i, col in enumerate(cols):
        obs.matrix_paste_column_in_place(col, i)

    d2 = maha_dist_sq_cpp(obs, flex.double(center), cov)
    return d2


def mcd_finite_sample(p, n, alpha):
    """Finite sample correction factor for the MCD estimate. Described in
    Pison et al. Metrika (2002). doi.org/10.1007/s001840200191. Implementation
    based on 'rawcorfactor' in fastmcd.m from Continuous Sound and Vibration
    Analysis by Edward Zechmann"""

    from scitbx.lstbx import normal_eqns

    if p > 2:
        coeffqpkwad500 = [
            [-1.42764571687802, 1.26263336932151, 2],
            [-1.06141115981725, 1.28907991440387, 3],
        ]
        coeffqpkwad875 = [
            [-0.455179464070565, 1.11192541278794, 2],
            [-0.294241208320834, 1.09649329149811, 3],
        ]

        y_500 = [
            math.log(-coeffqpkwad500[0][0] / p ** coeffqpkwad500[0][1]),
            math.log(-coeffqpkwad500[1][0] / p ** coeffqpkwad500[1][1]),
        ]
        y_875 = [
            math.log(-coeffqpkwad875[0][0] / p ** coeffqpkwad875[0][1]),
            math.log(-coeffqpkwad875[1][0] / p ** coeffqpkwad875[1][1]),
        ]
        A_500 = [
            [1, -math.log(coeffqpkwad500[0][2] * p**2)],
            [1, -math.log(coeffqpkwad500[1][2] * p**2)],
        ]
        A_875 = [
            [1, -math.log(coeffqpkwad875[0][2] * p**2)],
            [1, -math.log(coeffqpkwad875[1][2] * p**2)],
        ]

        # solve the set of equations labelled _500
        eqs = normal_eqns.linear_ls(2)
        for i in range(2):
            eqs.add_equation(
                right_hand_side=y_500[i],
                design_matrix_row=flex.double(A_500[i]),
                weight=1,
            )
        eqs.solve()
        coeffic_500 = eqs.solution()

        # solve the set of equations labelled _875
        eqs = normal_eqns.linear_ls(2)
        for i in range(2):
            eqs.add_equation(
                right_hand_side=y_875[i],
                design_matrix_row=flex.double(A_875[i]),
                weight=1,
            )
        eqs.solve()
        coeffic_875 = eqs.solution()

        fp_500_n = 1 - math.exp(coeffic_500[0]) / n ** coeffic_500[1]
        fp_875_n = 1 - math.exp(coeffic_875[0]) / n ** coeffic_875[1]

    elif p == 2:
        fp_500_n = 1 - math.exp(0.673292623522027) / n**0.691365864961895
        fp_875_n = 1 - math.exp(0.446537815635445) / n**1.06690782995919

    elif p == 1:
        fp_500_n = 1 - math.exp(0.262024211897096) / n**0.604756680630497
        fp_875_n = 1 - math.exp(-0.351584646688712) / n**1.01646567502486

    if alpha <= 0.875:
        fp_alpha_n = fp_500_n + (fp_875_n - fp_500_n) / 0.375 * (alpha - 0.5)

    if 0.875 < alpha and alpha <= 1:
        fp_alpha_n = fp_875_n + (1 - fp_875_n) / 0.125 * (alpha - 0.875)

    return 1 / fp_alpha_n


class FastMCD:
    """Experimental implementation of the FAST-MCD algorithm of Rousseeuw and
    van Driessen"""

    def __init__(
        self,
        data,
        alpha=0.5,
        max_n_groups=5,
        min_group_size=300,
        n_trials=500,
        k1=2,
        k2=2,
        k3=100,
    ):
        """data expected to be a list of flex.double arrays of the same length,
        representing the vectors of observations in each dimension"""

        # the full dataset as separate vectors
        self._data = data

        # number of variables
        self._p = len(self._data)
        # p == 1 is the univariate case, best dealt with using a different (exact)
        # algorithm. TODO: implement that here too.
        assert self._p > 1, "FastMCD init: Length of data must be > 1"

        # number of observations
        lens = [len(e) for e in self._data]
        assert all(
            e == lens[0] for e in lens
        ), "FastMCD init: Unequal length data arrays"
        self._n = lens[0]

        # some input checks
        assert (
            self._n > self._p
        ), f"FastMCD init: assert self._n > self._p ({self._n} > {self._p})"

        # default initial subset size
        self._alpha = alpha
        n2 = (self._n + self._p + 1) // 2
        self._h = int(math.floor(2 * n2 - self._n + 2 * (self._n - n2) * alpha))
        # In the original FAST-MCD, if h == n it reports a single
        # location and scatter estimate for the whole dataset and stops. Currently
        # limit this implementation to h < n
        assert (
            self._h < self._n
        ), f"FastMCD init: assert self._h < self._n ({self._h} < {self._n})"

        # set sizes of groups and how many trials to perform in each
        self._max_n_groups = max_n_groups
        self._min_group_size = min_group_size
        self._n_trials = n_trials

        # iteration limits
        self._k1 = k1
        self._k2 = k2
        self._k3 = k3

        # correction factors
        self._consistency_fac = mcd_consistency(self._p, self._h / self._n)
        self._finite_samp_fac = mcd_finite_sample(self._p, self._n, self._alpha)

        # perform calculation
        self._T_raw = None
        self._S_raw = None

        self.run()

    def run(self):
        """Run the Fast MCD calculation"""

        # algorithm for a small number of observations (up to twice the minimum
        # group size)
        if self._n < 2 * self._min_group_size:
            self._T_raw, self._S_raw = self.small_dataset_estimate()

        # algorithm for a larger number of observations
        else:
            self._T_raw, self._S_raw = self.large_dataset_estimate()

    def get_raw_T_and_S(self):
        """Get the raw MCD location (T) and covariance matrix (S) estimates"""

        return self._T_raw, self._S_raw

    def get_corrected_T_and_S(self):
        """Get the MCD location (T) and covariance matrix (S) estimates corrected
        for normal model consistency and finite-sample size"""

        fac = self._consistency_fac * self._finite_samp_fac
        return self._T_raw, self._S_raw * fac

    @staticmethod
    def means_and_covariance(vecs):
        """Prepare a dataset of equal length vectors for Mahalanobis distance
        squared calculation. The maha_dist_sq function requires the vectors,
        the vector of their means and their covariance matrix. Given the vectors,
        return the latter pair as a tuple"""

        center = flex.double([flex.mean(e) for e in vecs])
        covmat = cov(*vecs)
        return (center, covmat)

    @staticmethod
    def sample_data(data, sample_size):
        """sample (without replacement) the data vectors to select the same
        sample_size rows from each."""

        n = len(data[0])
        rows = flex.random_selection(n, sample_size)
        cols = [e.select(rows) for e in data]
        return cols

    def split_into_groups(self, sample, ngroups):
        """Split each vector in the data sample into groups of approximately equal
        size."""

        # number of obs in the sample
        sample_size = len(sample[0])

        # random permutation
        p = flex.random_permutation(sample_size)
        permuted = [col.select(p) for col in sample]

        # determine groups
        blocksize = int(sample_size / ngroups)
        rem = sample_size % ngroups
        blocksizes = [blocksize] * (ngroups - rem) + [blocksize + 1] * rem

        starts = [0]
        ends = [blocksizes[0]]
        for b in blocksizes[1:]:
            starts.append(ends[-1])
            ends.append(ends[-1] + b)
        blocks = zip(starts, ends)

        # split into groups
        groups = [[col[start:end] for col in permuted] for start, end in blocks]

        return groups

    def form_initial_subset(self, h, data):
        """Method 2 of subsection 3.1 of R&vD"""

        # permutation of input data for sampling
        p = flex.random_permutation(len(data[0]))
        permuted = [col.select(p) for col in data]

        # draw random p+1 subset J (or larger if required)
        detS0 = 0.0
        i = 0
        while not detS0 > 0.0:
            subset_size = self._p + 1 + i
            J = [e[0:subset_size] for e in permuted]
            i += 1
            T0, S0 = self.means_and_covariance(J)
            detS0 = S0.matrix_determinant_via_lu()

        H1 = self.concentration_step(h, data, T0, S0)
        return H1

    @staticmethod
    def concentration_step(h, data, T, S):
        """Practical application of Theorem 1 of R&vD"""

        d2s = maha_dist_sq(data, T, S)
        p = flex.sort_permutation(d2s)
        H1 = [col.select(p)[0:h] for col in data]
        return H1

    def small_dataset_estimate(self):
        """When a dataset is small, perform the initial trials directly on the
        whole dataset"""

        trials = []
        for i in range(self._n_trials):

            H1 = self.form_initial_subset(h=self._h, data=self._data)
            T1, S1 = self.means_and_covariance(H1)
            detS1 = S1.matrix_determinant_via_lu()

            # perform concentration steps
            detScurr, Tcurr, Scurr = detS1, T1, S1
            for j in range(self._k1):  # take maximum of k1 steps

                Hnew = self.concentration_step(self._h, self._data, Tcurr, Scurr)
                Tnew, Snew = self.means_and_covariance(Hnew)
                detSnew = Snew.matrix_determinant_via_lu()

                # detS3 < detS2 < detS1 by Theorem 1. In practice (rounding errors?)
                # this is not always the case here. Ensure that detScurr is no smaller than
                # one billionth the value of detSnew less than detSnew
                assert detScurr > (detSnew - detSnew / 1.0e9)
                detScurr, Tcurr, Scurr = detSnew, Tnew, Snew

            trials.append((detSnew, Tnew, Snew))

        # choose 10 trials with the lowest detS3
        trials.sort(key=lambda x: x[0])
        best_trials = []
        for i in range(10):
            detCurr, Tcurr, Scurr = trials[i]
            for j in range(self._k3):  # take maximum of k3 steps
                Hnew = self.concentration_step(self._h, self._data, Tcurr, Scurr)
                Tnew, Snew = self.means_and_covariance(Hnew)
                detNew = Snew.matrix_determinant_via_lu()
                if detNew == detCurr:
                    # print "trial {0}; iteration {1}; convergence".format(i,j)
                    break
                detCurr, Tcurr, Scurr = detNew, Tnew, Snew
            best_trials.append((detNew, Tnew, Snew))

        # Find the minimum covariance determinant from that set of 10
        best_trials.sort(key=lambda x: x[0])
        _, Tbest, Sbest = best_trials[0]
        return Tbest, Sbest

    def large_dataset_estimate(self):
        """When a dataset is large, construct disjoint subsets of the full data
        and perform initial trials within each of these, then merge"""

        ngroups = int(self._n / self._min_group_size)
        if ngroups < self._max_n_groups:
            # use all the data and split into ngroups
            sample_size = self._n
        else:
            # sample the data and split into the maximum number of groups
            ngroups = self._max_n_groups
            sample_size = self._min_group_size * self._max_n_groups

        # sample the data and split into groups
        sampled = self.sample_data(self._data, sample_size=sample_size)
        groups = self.split_into_groups(sample=sampled, ngroups=ngroups)

        # work within the groups now
        n_trials = self._n_trials // ngroups
        trials = []
        h_frac = self._h / self._n
        for group in groups:

            h_sub = int(len(group[0]) * h_frac)
            gp_trials = []
            for i in range(n_trials):

                H1 = self.form_initial_subset(h=h_sub, data=group)
                T1, S1 = self.means_and_covariance(H1)
                detS1 = S1.matrix_determinant_via_lu()

                # perform concentration steps
                detScurr, Tcurr, Scurr = detS1, T1, S1
                for j in range(self._k1):  # take k1 steps

                    Hnew = self.concentration_step(h_sub, group, Tcurr, Scurr)
                    Tnew, Snew = self.means_and_covariance(Hnew)
                    detSnew = Snew.matrix_determinant_via_lu()

                    # detS3 < detS2 < detS1 by Theorem 1. In practice (rounding errors?)
                    # this is not always the case here. Ensure that detScurr is no smaller than
                    # one billionth the value of detSnew less than detSnew
                    assert detScurr > (detSnew - detSnew / 1.0e9)
                    detScurr, Tcurr, Scurr = detSnew, Tnew, Snew

                gp_trials.append((detSnew, Tnew, Snew))

            # choose 10 trials with the lowest determinant and put in the outer list
            gp_trials.sort(key=lambda x: x[0])
            trials.extend(gp_trials[0:10])

        # now have 10 best trials from each group. Work with the merged (==sampled)
        # set
        mrgd_trials = []
        h_mrgd = int(sample_size * h_frac)
        for trial in trials:

            detScurr, Tcurr, Scurr = trial
            for j in range(self._k2):  # take k2 steps

                Hnew = self.concentration_step(h_mrgd, sampled, Tcurr, Scurr)
                Tnew, Snew = self.means_and_covariance(Hnew)
                detSnew = Snew.matrix_determinant_via_lu()
                detScurr, Tcurr, Scurr = detSnew, Tnew, Snew

            mrgd_trials.append((detSnew, Tnew, Snew))

        # sort trials by the lowest detS3 and work with the whole dataset now
        mrgd_trials.sort(key=lambda x: x[0])

        # choose number of steps to iterate based on dataset size (ugly)
        size = self._n * self._p
        if size <= 100000:
            k4 = self._k3
        elif size <= 200000:
            k4 = 10
        elif size <= 300000:
            k4 = 9
        elif size <= 400000:
            k4 = 8
        elif size <= 500000:
            k4 = 7
        elif size <= 600000:
            k4 = 6
        elif size <= 700000:
            k4 = 5
        elif size <= 800000:
            k4 = 4
        elif size <= 900000:
            k4 = 3
        elif size <= 1000000:
            k4 = 2
        else:
            k4 = 1

        # choose number of trials to look at based on number of obs (ugly)
        n_reps = 1 if self._n > 5000 else 10

        best_trials = []
        for i in range(n_reps):
            detCurr, Tcurr, Scurr = mrgd_trials[i]
            for j in range(k4):  # take maximum of k4 steps
                Hnew = self.concentration_step(self._h, self._data, Tcurr, Scurr)
                Tnew, Snew = self.means_and_covariance(Hnew)
                detNew = Snew.matrix_determinant_via_lu()
                if detNew == detCurr:
                    # print "trial {0}; iteration {1}; convergence".format(i,j)
                    break
                detCurr, Tcurr, Scurr = detNew, Tnew, Snew
            # print "determinant", detNew
            # print "location:", list(Tnew)
            # print "covariance:"
            # print Snew.as_scitbx_matrix()
            best_trials.append((detNew, Tnew, Snew))

        # Find the minimum covariance determinant from that set of 10
        best_trials.sort(key=lambda x: x[0])
        _, Tbest, Sbest = best_trials[0]
        return Tbest, Sbest



from __future__ import annotations

import collections
import copy
import glob
import logging
import os
import pickle
import sys
import tarfile
import time
from io import BytesIO

from dxtbx.model.experiment_list import (
    Experiment,
    ExperimentList,
    ExperimentListFactory,
)
from libtbx.phil import parse
from libtbx.utils import Abort, Sorry

import dials.util
from dials.array_family import flex
from dials.util import log

logger = logging.getLogger("dials.command_line.stills_process")

help_message = """
DIALS script for processing still images. Import, index, refine, and integrate are all done for each image
separately.
"""


def _control_phil_str():
    return """
  input {
    file_list = None
      .type = path
      .help = Path to a text file with a list of images
    glob = None
      .type = str
      .help = For large, multi-file datasets, specify the paths using wildcards (e.g. *.cbf)
      .multiple = True
    image_tag = None
      .type = str
      .multiple = True
      .help = Only process images with these tag(s). For single-image files (like CBFs or SMVs), the image \
              tag for each file is the file name. For multi-image files like HDF5, the image tag is        \
              filename_imagenumber (including leading zeros). Use show_image_tags=True to see the list     \
              of image tags that will be used for a dataset.
    show_image_tags = False
      .type = bool
      .help = Show the set of image tags that would be used during processing. To process subsets of image \
              files, use these tags with the image_tag parameter.
    max_images = None
      .type = int
      .help = Limit total number of processed images to max_images
    ignore_gain_mismatch = False
      .type = bool
      .expert_level = 3
      .help = Detector gain should be set on the detector models loaded from the images or in the \
              processing parameters, not both. Override the check that this is true with this flag. \
  }

  dispatch {
    pre_import = False
      .type = bool
      .expert_level = 2
      .help = If True, before processing import all the data. Needed only if processing \
              multiple multi-image files at once (not a recommended use case)
    process_percent = None
      .type = int(value_min=1, value_max=100)
      .help = Percent of events to process
    refine = False
      .expert_level = 2
      .type = bool
      .help = If True, after indexing, refine the experimental models
    squash_errors = True
      .expert_level = 2
      .type = bool
      .help = If True, if an image fails to process, continue to the next image. \
              otherwise, halt processing and show the error.
    find_spots = True
      .expert_level = 2
      .type = bool
      .help = Whether to do spotfinding. Needed for indexing/integration
    index = True
      .expert_level = 2
      .type = bool
      .help = Attempt to index images. find_spots also needs to be True for this to work
    integrate = True
      .expert_level = 2
      .type = bool
      .help = Integrate indexed images. Ignored if index=False or find_spots=False
    coset = False
      .expert_level = 2
      .type = bool
      .help = Within the integrate dispatcher, integrate a sublattice coset intended to represent \
              negative control spots with no Bragg diffraction.
    hit_finder{
      enable = True
        .type = bool
        .help = Whether to do hitfinding. hit_finder=False: process all images
      minimum_number_of_reflections = 16
        .type = int
        .help = If the number of strong reflections on an image is less than this, and \
                 the hitfinder is enabled, discard this image.
      maximum_number_of_reflections = None
       .type = int
       .help = If specified, ignores images with more than this many number of reflections
    }
  }

  output {
    output_dir = .
      .type = str
      .help = Directory output files will be placed
    composite_output = True
      .type = bool
      .help = If True, save one set of experiment/reflection files per process, where each is a \
              concatenated list of all the successful events examined by that process. \
              If False, output a separate experiment/reflection file per image (generates a \
              lot of files).
    logging_dir = None
      .type = str
      .help = Directory output log files will be placed
    experiments_filename = None
      .type = str
      .help = The filename for output experiments. For example, %s_imported.expt
    strong_filename = None
      .type = str
      .help = The filename for strong reflections from spot finder output. For example: \
              %s_strong.refl
    indexed_filename = %s_indexed.refl
      .type = str
      .help = The filename for indexed reflections.
    refined_experiments_filename = %s_refined.expt
      .type = str
      .help = The filename for saving refined experimental models
    integrated_filename = %s_integrated.refl
      .type = str
      .help = The filename for final integrated reflections.
    integrated_experiments_filename = %s_integrated.expt
      .type = str
      .help = The filename for saving final experimental models.
    coset_filename = %s_coset%d.refl
      .type = str
      .help = The filename for final coset reflections.
    coset_experiments_filename = %s_coset%d.expt
      .type = str
      .help = The filename for saving final coset experimental models.
    profile_filename = None
      .type = str
      .help = The filename for output reflection profile parameters
    integration_pickle = int-%d-%s.pickle
      .type = str
      .help = Filename for cctbx.xfel-style integration pickle files
  }

  mp {
    method = *multiprocessing sge lsf pbs mpi
      .type = choice
      .help = "The multiprocessing method to use"
    nproc = 1
      .type = int(value_min=1)
      .help = "The number of processes to use."
    composite_stride = None
      .type = int
      .help = For MPI, if using composite mode, specify how many ranks to    \
              aggregate data from.  For example, if you have 100 processes,  \
              composite mode will output N*100 files, where N is the number  \
              of file types (expt, refl, etc). If you specify stride = 25, \
              then each group of 25 process will send their results to 4     \
              processes and only N*4 files will be created. Ideally, match   \
              stride to the number of processors per node.
    debug
      .expert_level = 2
    {
      cProfile = False
        .type = bool
        .help = Enable code profiling. Profiling file will be available in  \
                the debug folder. Use (for example) runsnake to visualize   \
                processing performance
      output_debug_logs = True
        .type = bool
        .help = Whether to write debugging information for every image      \
                processed
    }
  }
"""


def _dials_phil_str():
    return """
  input {
    reference_geometry = None
      .type = str
      .help = Provide an models.expt file with exactly one detector model. Data processing will use \
              that geometry instead of the geometry found in the image headers.
    sync_reference_geom = True
      .type = bool
      .help = ensures the reference hierarchy agrees with the image format
  }

  output {
    shoeboxes = True
      .type = bool
      .help = Save the raw pixel values inside the reflection shoeboxes during spotfinding.
  }

  include scope dials.util.options.geometry_phil_scope
  include scope dials.algorithms.spot_finding.factory.phil_scope
  include scope dials.algorithms.indexing.indexer.phil_scope
  indexing {
      include scope dials.algorithms.indexing.lattice_search.basis_vector_search_phil_scope
  }
  include scope dials.algorithms.refinement.refiner.phil_scope
  include scope dials.algorithms.integration.integrator.phil_scope
  include scope dials.algorithms.profile_model.factory.phil_scope
  include scope dials.algorithms.spot_prediction.reflection_predictor.phil_scope
  include scope dials.algorithms.integration.stills_significance_filter.phil_scope

  indexing {
    stills {
      method_list = None
        .type = strings
        .help = List of indexing methods. If indexing fails with first method, indexing will be \
                attempted with the next, and so forth
      known_orientations = None
        .type = path
        .multiple = True
        .expert_level = 2
        .help = Paths to previous processing results including crystal orientations. \
                If specified, images will not be re-indexed, but instead the known \
                orientations will be used. Provide paths to experiment list files, using \
                wildcards as needed.
      require_known_orientation = False
        .type = bool
        .expert_level = 2
        .help = If known_orientations are provided, and an orientation for an image is not \
                found, this is whether or not to attempt to index the image from scratch \
                using indexing.method
    }
  }

  integration {
    include scope dials.algorithms.integration.kapton_correction.absorption_phil_scope
    coset {
      transformation = 6
        .type = int(value_min=0, value_max=6)
        .multiple = False
        .help = The index number(s) of the modulus=2 sublattice transformation(s) used to produce distince coset results. \
                0=Double a, 1=Double b, 2=Double c, 3=C-face centering, 4=B-face centering, 5=A-face centering, 6=Body centering \
                See Sauter and Zwart, Acta D (2009) 65:553
    }

    integration_only_overrides {
      trusted_range = None
        .type = floats(size=2)
        .help = "Override the panel trusted range [min_trusted_value, max_trusted_value] during integration."
        .short_caption = "Panel trusted range"
    }
  }

  profile {
    gaussian_rs {
      parameters {
        sigma_b_cutoff = 0.1
          .type = float
          .help = Maximum sigma_b before the image is rejected
      }
    }
  }
"""


def _program_defaults_phil_str():
    return """
indexing {
  method = fft1d
}
refinement {
  parameterisation {
    auto_reduction {
      min_nref_per_parameter = 1
      action = fix
    }
    beam.fix = all
    detector.fix = all
  }
  reflections {
    weighting_strategy.override = stills
    outlier.algorithm = null
  }
}
integration {
  integrator = stills
  profile.fitting = False
  background {
    algorithm = simple
    simple {
      outlier.algorithm = plane
      model.algorithm = linear2d
    }
  }
}
profile.gaussian_rs.min_spots.overall = 0
"""


control_phil_str = _control_phil_str()
dials_phil_str = _dials_phil_str()
program_defaults_phil_str = _program_defaults_phil_str()

phil_scope = parse(control_phil_str + dials_phil_str, process_includes=True).fetch(
    parse(program_defaults_phil_str)
)


def do_import(filename, load_models=True):
    logger.info("Loading %s", os.path.basename(filename))
    experiments = ExperimentListFactory.from_filenames([filename], load_models=False)
    if len(experiments) == 0:
        try:
            experiments = ExperimentListFactory.from_json_file(filename)
        except ValueError:
            raise Abort(f"Could not load {filename}")

    if len(experiments) == 0:
        raise Abort(f"Could not load {filename}")

    from dxtbx.imageset import ImageSetFactory

    all_experiments = ExperimentList()
    for experiment in experiments:
        # Convert from ImageSequence to ImageSet, if needed
        imageset = ImageSetFactory.imageset_from_anyset(experiment.imageset)
        for i in range(len(imageset)):
            # Preserve original models if they were available (in the case of an image file
            # they will not be, but in the case of a previously processed experiment list,
            # then they may be available
            expt = Experiment(
                imageset=imageset[i : i + 1],
                detector=experiment.detector,
                beam=experiment.beam,
                scan=experiment.scan,
                goniometer=experiment.goniometer,
                crystal=experiment.crystal,
            )
            if load_models:
                expt.load_models()
            all_experiments.append(expt)

    return all_experiments


def sync_geometry(src, dest):
    dest.set_local_frame(
        src.get_local_fast_axis(), src.get_local_slow_axis(), src.get_local_origin()
    )
    if not src.is_panel():
        for src_child, dest_child in zip(src, dest):
            sync_geometry(src_child, dest_child)


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        # The script usage
        usage = "usage: dials.stills_process [options] [param.phil] filenames"

        self.tag = None
        self.reference_detector = None

        # Create the parser
        self.parser = ArgumentParser(usage=usage, phil=phil_scope, epilog=help_message)

    def load_reference_geometry(self):
        if self.params.input.reference_geometry is None:
            return

        try:
            ref_experiments = ExperimentListFactory.from_json_file(
                self.params.input.reference_geometry, check_format=False
            )
        except Exception:
            try:
                import dxtbx

                img = dxtbx.load(self.params.input.reference_geometry)
            except Exception:
                raise Sorry(
                    "Couldn't load geometry file %s"
                    % self.params.input.reference_geometry
                )
            else:
                self.reference_detector = img.get_detector()
        else:
            assert len(ref_experiments.detectors()) == 1
            self.reference_detector = ref_experiments.detectors()[0]

    def run(self, args=None):
        """Execute the script."""
        from libtbx import easy_mp

        try:
            from mpi4py import MPI
        except ImportError:
            rank = 0
            size = 1
        else:
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()  # each process in MPI has a unique id, 0-indexed
            size = comm.Get_size()  # size: number of processes running in this job

        if rank == 0:
            # Parse the command line
            params, options, all_paths = self.parser.parse_args(
                args, show_diff_phil=False, return_unhandled=True, quick_parse=True
            )

            if params.input.glob:
                all_paths.extend(params.input.glob)
            globbed = []
            for p in all_paths:
                g = glob.glob(p)
                if not g:
                    sys.exit(f"Error: Unhandled path or option: {p}")
                globbed.extend(g)
            all_paths = globbed

            if not all_paths and params.input.file_list is not None:
                all_paths.extend(
                    [path.strip() for path in open(params.input.file_list).readlines()]
                )

            if params.indexing.stills.known_orientations:
                known_orientations = {}
                for path in params.indexing.stills.known_orientations:
                    for g in glob.glob(path):
                        ko_expts = ExperimentList.from_file(g, check_format=False)
                        for expt in ko_expts:
                            assert (
                                len(expt.imageset.indices()) == 1
                                and len(expt.imageset.paths()) == 1
                            )
                            key = (
                                os.path.basename(expt.imageset.paths()[0]),
                                expt.imageset.indices()[0],
                            )
                            if key not in known_orientations:
                                known_orientations[key] = []
                            known_orientations[key].append(expt.crystal)
                if not known_orientations:
                    raise Sorry(
                        "No known_orientations found at the locations specified: %s"
                        % ", ".join(params.indexing.stills.known_orientations)
                    )
                params.indexing.stills.known_orientations = known_orientations
        if size > 1:
            if rank == 0:
                transmitted_info = params, options, all_paths
            else:
                transmitted_info = None
            params, options, all_paths = comm.bcast(transmitted_info, root=0)

        # Check we have some filenames
        if not all_paths:
            self.parser.print_help()
            return

        if params.mp.debug.cProfile:
            import cProfile

            self.pr = cProfile.Profile()
            self.pr.enable()

        print(f"Have {len(all_paths)} files")

        # Mask validation
        for mask_path in params.spotfinder.lookup.mask, params.integration.lookup.mask:
            if mask_path is not None and not os.path.isfile(mask_path):
                raise Sorry(f"Mask {mask_path} not found")

        # Save the options
        self.options = options
        self.params = params

        st = time.time()

        if params.mp.method == "mpi":
            # Configure the logging
            if params.output.logging_dir is None:
                logfile = None
            else:
                log_path = os.path.join(
                    params.output.logging_dir, "log_rank%04d.out" % rank
                )
                error_path = os.path.join(
                    params.output.logging_dir, "error_rank%04d.out" % rank
                )
                print(f"Redirecting stdout to {log_path}")
                print(f"Redirecting stderr to {error_path}")
                sys.stdout = open(log_path, "a")
                sys.stderr = open(error_path, "a")
                print("Should be redirected now")

                logfile = os.path.join(
                    params.output.logging_dir, "info_rank%04d.out" % rank
                )

            log.config(verbosity=options.verbose, logfile=logfile)

        else:

            # Configure logging
            log.config(verbosity=options.verbose, logfile="dials.process.log")

        bad_phils = [f for f in all_paths if os.path.splitext(f)[1] == ".phil"]
        if len(bad_phils) > 0:
            self.parser.print_help()
            logger.error(
                "Error: the following phil files were not understood: %s",
                ", ".join(bad_phils),
            )
            return

        # Log the diff phil
        diff_phil = self.parser.diff_phil.as_str()
        if diff_phil != "":
            logger.info("The following parameters have been modified:\n")
            logger.info(diff_phil)

        for abs_params in self.params.integration.absorption_correction:
            if abs_params.apply:
                if not (
                    self.params.integration.debug.output
                    and not self.params.integration.debug.separate_files
                ):
                    raise Sorry(
                        "Shoeboxes must be saved to integration intermediates to apply an absorption correction. "
                        + "Set integration.debug.output=True, integration.debug.separate_files=False and "
                        + "integration.debug.delete_shoeboxes=True to temporarily store shoeboxes."
                    )

        self.load_reference_geometry()
        from dials.command_line.dials_import import ManualGeometryUpdater

        update_geometry = ManualGeometryUpdater(params)

        # Import stuff
        logger.info("Loading files...")
        pre_import = params.dispatch.pre_import or len(all_paths) == 1
        if pre_import:
            # Handle still imagesets by breaking them apart into multiple experiments
            # Further handle single file still imagesets (like HDF5) by tagging each
            # frame using its index

            experiments = ExperimentList()
            for path in sorted(all_paths):
                experiments.extend(do_import(path, load_models=False))

            indices = []
            basenames = []
            basename_counts = {}
            split_experiments = []
            for i, imageset in enumerate(experiments.imagesets()):
                assert len(imageset) == 1
                paths = imageset.paths()
                indices.append(i)
                basename = os.path.splitext(os.path.basename(paths[0]))[0]
                basenames.append(basename)
                if basename in basename_counts:
                    basename_counts[basename] += 1
                else:
                    basename_counts[basename] = 1
                split_experiments.append(experiments[i : i + 1])
            tags = []
            split_experiments2 = []
            for i, basename in zip(indices, basenames):
                if basename_counts[basename] > 1:
                    tag = "%s_%05d" % (basename, i)
                else:
                    tag = basename
                if (
                    not self.params.input.image_tag
                    or tag in self.params.input.image_tag
                ):
                    tags.append(tag)
                    split_experiments2.append(split_experiments[i])
            split_experiments = split_experiments2

            # Wrapper function
            def do_work(i, item_list, processor=None, finalize=True):
                if not processor:
                    processor = Processor(
                        copy.deepcopy(params), composite_tag="%04d" % i, rank=i
                    )

                for item in item_list:
                    tag = item[0]
                    experiments = split_experiments[item[1]]
                    try:
                        assert len(experiments) == 1
                        experiment = experiments[0]
                        experiment.load_models()
                        imageset = experiment.imageset
                        update_geometry(imageset)
                        experiment.beam = imageset.get_beam()
                        experiment.detector = imageset.get_detector()
                    except RuntimeError as e:
                        logger.warning("Error updating geometry on item %s, %s", tag, e)
                        continue

                    if self.reference_detector is not None:
                        experiment = experiments[0]
                        if self.params.input.sync_reference_geom:
                            imageset = experiment.imageset
                            sync_geometry(
                                self.reference_detector.hierarchy(),
                                imageset.get_detector().hierarchy(),
                            )
                            experiment.detector = imageset.get_detector()
                        else:
                            experiment.detector = copy.deepcopy(self.reference_detector)

                    processor.process_experiments(tag, experiments)
                    imageset.clear_cache()
                if finalize:
                    processor.finalize()
                return processor

            iterable = list(zip(tags, range(len(split_experiments))))

        else:
            basenames = collections.defaultdict(int)
            sorted_paths = sorted(all_paths)
            for filename in sorted_paths:
                basename = os.path.splitext(os.path.basename(filename))[0]
                basenames[basename] += 1
            tags = []
            all_paths2 = []
            for i, (basename, count) in enumerate(basenames.items()):
                if count > 1:
                    tag = "%s_%05d" % (basename, i)
                else:
                    tag = basename
                if (
                    not self.params.input.image_tag
                    or tag in self.params.input.image_tag
                ):
                    tags.append(tag)
                    all_paths2.append(sorted_paths[i])
            all_paths = all_paths2

            # Wrapper function
            def do_work(i, item_list, processor=None, finalize=True):
                if not processor:
                    processor = Processor(
                        copy.deepcopy(params), composite_tag="%04d" % i, rank=i
                    )
                for item in item_list:
                    tag, filename = item

                    experiments = do_import(filename, load_models=True)
                    imagesets = experiments.imagesets()
                    if len(imagesets) == 0 or len(imagesets[0]) == 0:
                        logger.info("Zero length imageset in file: %s", filename)
                        return
                    if len(imagesets) > 1:
                        raise Abort(f"Found more than one imageset in file: {filename}")
                    if len(imagesets[0]) > 1:
                        raise Abort(
                            "Found a multi-image file. Run again with pre_import=True"
                        )

                    try:
                        update_geometry(imagesets[0])
                        experiment = experiments[0]
                        experiment.beam = imagesets[0].get_beam()
                        experiment.detector = imagesets[0].get_detector()
                    except RuntimeError as e:
                        logger.warning("Error updating geometry on item %s, %s", tag, e)
                        continue

                    if self.reference_detector is not None:
                        if self.params.input.sync_reference_geom:
                            imageset = experiments[0].imageset
                            sync_geometry(
                                self.reference_detector.hierarchy(),
                                imageset.get_detector().hierarchy(),
                            )
                            experiments[0].detector = imageset.get_detector()
                        else:
                            experiments[0].detector = copy.deepcopy(
                                self.reference_detector
                            )

                    processor.process_experiments(tag, experiments)
                if finalize:
                    processor.finalize()
                return processor

            iterable = list(zip(tags, all_paths))

        if params.input.max_images:
            iterable = iterable[: params.input.max_images]

        if params.input.show_image_tags:
            print("Showing image tags for this dataset and exiting")
            for tag, item in iterable:
                print(tag)
            return

        # prepare fractions of process_percent, if given
        process_fractions = None
        if params.dispatch.process_percent:
            import fractions

            percent = params.dispatch.process_percent / 100
            process_fractions = fractions.Fraction(percent).limit_denominator(100)

            def process_this_event(nevent):
                # nevent modulo the denominator gives us which fraction we're in
                n_mod_denom = nevent % process_fractions.denominator
                # compare the 0-indexed modulo against the 1-indexed numerator (intentionally not <=)
                n_accept = n_mod_denom < process_fractions.numerator
                return n_accept

        # Process the data
        if params.mp.method == "mpi":
            if size <= 2:  # client/server only makes sense for n>2
                subset = [
                    item for i, item in enumerate(iterable) if (i + rank) % size == 0
                ]
                do_work(rank, subset)
            else:
                processor = Processor(
                    copy.deepcopy(params), composite_tag="%04d" % rank, rank=rank
                )

                if rank == 0:
                    # server process
                    for item_num, item in enumerate(iterable):
                        if process_fractions and not process_this_event(item_num):
                            continue

                        print("Getting next available process")
                        rankreq = comm.recv(source=MPI.ANY_SOURCE)
                        print(f"Process {rankreq} is ready, sending {item[0]}\n")
                        comm.send(item, dest=rankreq)
                    # send a stop command to each process
                    print("MPI DONE, sending stops\n")
                    for rankreq in range(size - 1):
                        rankreq = comm.recv(source=MPI.ANY_SOURCE)
                        print("Sending stop to %d\n" % rankreq)
                        comm.send("endrun", dest=rankreq)
                    print("All stops sent.")

                else:
                    # client process
                    while True:
                        # inform the server this process is ready for an event
                        print("Rank %d getting next task" % rank)
                        comm.send(rank, dest=0)
                        print("Rank %d waiting for response" % rank)
                        item = comm.recv(source=0)
                        if item == "endrun":
                            print("Rank %d received endrun" % rank)
                            break
                        print("Rank %d beginning processing" % rank)
                        try:
                            processor = do_work(rank, [item], processor, finalize=False)
                        except Exception as e:
                            print(
                                "Rank %d unhandled exception processing event" % rank,
                                str(e),
                            )
                        print("Rank %d event processed" % rank)
                processor.finalize()
        else:
            from dxtbx.command_line.image_average import splitit

            if params.mp.nproc == 1:
                do_work(0, iterable)
            else:
                result = list(
                    easy_mp.multi_core_run(
                        myfunction=do_work,
                        argstuples=list(enumerate(splitit(iterable, params.mp.nproc))),
                        nproc=params.mp.nproc,
                    )
                )
                error_list = [r[2] for r in result]
                if error_list.count(None) != len(error_list):
                    print(
                        "Some processes failed execution. Not all images may have processed. Error messages:"
                    )
                    for error in error_list:
                        if error is None:
                            continue
                        print(error)

        # Total Time
        logger.info("")
        logger.info("Total Time Taken = %f seconds", time.time() - st)

        if params.mp.debug.cProfile:
            self.pr.disable()
            self.pr.dump_stats(
                os.path.join(
                    self.params.output.output_dir, "debug", "cpu_%d.prof" % comm.rank
                )
            )


class Processor:
    def __init__(self, params, composite_tag=None, rank=0):
        self.params = params
        self.composite_tag = composite_tag

        # The convention is to put %s in the phil parameter to add a tag to
        # each output datafile. Save the initial templates here.
        self.experiments_filename_template = params.output.experiments_filename
        self.strong_filename_template = params.output.strong_filename
        self.indexed_filename_template = params.output.indexed_filename
        self.refined_experiments_filename_template = (
            params.output.refined_experiments_filename
        )
        self.integrated_filename_template = params.output.integrated_filename
        self.integrated_experiments_filename_template = (
            params.output.integrated_experiments_filename
        )
        if params.dispatch.coset:
            self.coset_filename_template = params.output.coset_filename
            self.coset_experiments_filename_template = (
                params.output.coset_experiments_filename
            )

        debug_dir = os.path.join(params.output.output_dir, "debug")
        if not os.path.exists(debug_dir):
            try:
                os.makedirs(debug_dir)
            except OSError:
                pass  # due to multiprocessing, makedirs can sometimes fail
        assert os.path.exists(debug_dir)
        self.debug_file_path = os.path.join(debug_dir, "debug_%d.txt" % rank)
        write_newline = os.path.exists(self.debug_file_path)
        if write_newline:  # needed if the there was a crash
            self.debug_write("")

        if params.output.composite_output:
            assert composite_tag is not None

            self.all_imported_experiments = ExperimentList()
            self.all_strong_reflections = flex.reflection_table()
            self.all_indexed_experiments = ExperimentList()
            self.all_indexed_reflections = flex.reflection_table()
            self.all_integrated_experiments = ExperimentList()
            self.all_integrated_reflections = flex.reflection_table()
            self.all_int_pickle_filenames = []
            self.all_int_pickles = []
            self.all_coset_experiments = ExperimentList()
            self.all_coset_reflections = flex.reflection_table()

            self.setup_filenames(composite_tag)

    def setup_filenames(self, tag):
        # before processing, set output paths according to the templates
        if (
            self.experiments_filename_template is not None
            and "%s" in self.experiments_filename_template
        ):
            self.params.output.experiments_filename = os.path.join(
                self.params.output.output_dir,
                self.experiments_filename_template % ("idx-" + tag),
            )
        if (
            self.strong_filename_template is not None
            and "%s" in self.strong_filename_template
        ):
            self.params.output.strong_filename = os.path.join(
                self.params.output.output_dir,
                self.strong_filename_template % ("idx-" + tag),
            )
        if (
            self.indexed_filename_template is not None
            and "%s" in self.indexed_filename_template
        ):
            self.params.output.indexed_filename = os.path.join(
                self.params.output.output_dir,
                self.indexed_filename_template % ("idx-" + tag),
            )
        if (
            self.refined_experiments_filename_template is not None
            and "%s" in self.refined_experiments_filename_template
        ):
            self.params.output.refined_experiments_filename = os.path.join(
                self.params.output.output_dir,
                self.refined_experiments_filename_template % ("idx-" + tag),
            )
        if (
            self.integrated_filename_template is not None
            and "%s" in self.integrated_filename_template
        ):
            self.params.output.integrated_filename = os.path.join(
                self.params.output.output_dir,
                self.integrated_filename_template % ("idx-" + tag),
            )
        if (
            self.integrated_experiments_filename_template is not None
            and "%s" in self.integrated_experiments_filename_template
        ):
            self.params.output.integrated_experiments_filename = os.path.join(
                self.params.output.output_dir,
                self.integrated_experiments_filename_template % ("idx-" + tag),
            )
        if (
            self.params.dispatch.coset
            and self.coset_filename_template is not None
            and "%s" in self.coset_filename_template
        ):
            self.params.output.coset_filename = os.path.join(
                self.params.output.output_dir,
                self.coset_filename_template
                % ("idx-" + tag, self.params.integration.coset.transformation),
            )
        if (
            self.params.dispatch.coset
            and self.coset_experiments_filename_template is not None
            and "%s" in self.coset_experiments_filename_template
        ):
            self.params.output.coset_experiments_filename = os.path.join(
                self.params.output.output_dir,
                self.coset_experiments_filename_template
                % ("idx-" + tag, self.params.integration.coset.transformation),
            )

    def debug_start(self, tag):
        if not self.params.mp.debug.output_debug_logs:
            return

        import socket

        self.debug_str = f"{socket.gethostname()},{tag}"
        self.debug_str += ",%s,%s,%s\n"
        self.debug_write("start")

    def debug_write(self, string, state=None):
        if not self.params.mp.debug.output_debug_logs:
            return

        from xfel.cxi.cspad_ana import cspad_tbx  # XXX move to common timestamp format

        ts = cspad_tbx.evt_timestamp()  # Now
        debug_file_handle = open(self.debug_file_path, "a")
        if string == "":
            debug_file_handle.write("\n")
        else:
            if state is None:
                state = "    "
            debug_file_handle.write(self.debug_str % (ts, state, string))
        debug_file_handle.close()

    def process_experiments(self, tag, experiments):

        if not self.params.output.composite_output:
            self.setup_filenames(tag)
        self.tag = tag
        self.debug_start(tag)

        if self.params.output.experiments_filename:
            if self.params.output.composite_output:
                self.all_imported_experiments.extend(experiments)
            else:
                experiments.as_json(self.params.output.experiments_filename)

        # Do the processing
        try:
            self.pre_process(experiments)
        except Exception as e:
            print("Error in pre-process", tag, str(e))
            self.debug_write("preprocess_exception", "fail")
            if not self.params.dispatch.squash_errors:
                raise
            return
        try:
            if self.params.dispatch.find_spots:
                self.debug_write("spotfind_start")
                observed = self.find_spots(experiments)
            else:
                print("Spot Finding turned off. Exiting")
                self.debug_write("data_loaded", "done")
                return
        except Exception as e:
            print("Error spotfinding", tag, str(e))
            self.debug_write("spotfinding_exception", "fail")
            if not self.params.dispatch.squash_errors:
                raise
            return
        try:
            if self.params.dispatch.index:
                if (
                    self.params.dispatch.hit_finder.enable
                    and len(observed)
                    < self.params.dispatch.hit_finder.minimum_number_of_reflections
                ):
                    print("Not enough spots to index", tag)
                    self.debug_write(f"not_enough_spots_{len(observed)}", "stop")
                    return
                if (
                    self.params.dispatch.hit_finder.maximum_number_of_reflections
                    is not None
                ):
                    if (
                        self.params.dispatch.hit_finder.enable
                        and len(observed)
                        > self.params.dispatch.hit_finder.maximum_number_of_reflections
                    ):
                        print("Too many spots to index - Possibly junk", tag)
                        self.debug_write(f"too_many_spots_{len(observed)}", "stop")
                        return
                self.debug_write("index_start")
                experiments, indexed = self.index(experiments, observed)
            else:
                print("Indexing turned off. Exiting")
                self.debug_write(f"spotfinding_ok_{len(observed)}", "done")
                return
        except Exception as e:
            print("Couldn't index", tag, str(e))
            if not self.params.dispatch.squash_errors:
                raise
            self.debug_write(f"indexing_failed_{len(observed)}", "stop")
            return
        self.debug_write("refine_start")
        try:
            experiments, indexed = self.refine(experiments, indexed)
        except Exception as e:
            print("Error refining", tag, str(e))
            self.debug_write(f"refine_failed_{len(indexed)}", "fail")
            if not self.params.dispatch.squash_errors:
                raise
            return
        try:
            if self.params.dispatch.integrate:
                self.debug_write("integrate_start")
                integrated = self.integrate(experiments, indexed)
            else:
                print("Integration turned off. Exiting")
                self.debug_write(f"index_ok_{len(indexed)}", "done")
                return
        except Exception as e:
            print("Error integrating", tag, str(e))
            self.debug_write(f"integrate_failed_{len(indexed)}", "fail")
            if not self.params.dispatch.squash_errors:
                raise
            return
        self.debug_write(f"integrate_ok_{len(integrated)}", "done")

    def pre_process(self, experiments):
        """Add any pre-processing steps here"""
        if (
            self.params.indexing.stills.known_orientations
            and self.params.indexing.stills.require_known_orientation
        ):
            for expt in experiments:
                assert (
                    len(expt.imageset.indices()) == 1
                    and len(expt.imageset.paths()) == 1
                )
                key = (
                    os.path.basename(expt.imageset.paths()[0]),
                    expt.imageset.indices()[0],
                )
                if key not in self.params.indexing.stills.known_orientations:
                    raise RuntimeError("Image not found in set of known orientations")

        if not self.params.input.ignore_gain_mismatch:
            g1 = self.params.spotfinder.threshold.dispersion.gain
            g2 = self.params.integration.summation.detector_gain
            gain = g1 if g1 is not None else g2
            if gain is not None and gain != 1.0:
                for detector in experiments.detectors():
                    for panel in detector:
                        if panel.get_gain() != 1.0 and panel.get_gain() != gain:
                            raise RuntimeError(
                                """
The detector is reporting a gain of %f but you have also supplied a gain of %f. Since the detector gain is not 1.0, your supplied gain will be multiplicatively applied in addition to the detector's gain, which is unlikely to be correct. Please re-run, removing spotfinder.dispersion.gain and integration.summation.detector_gain from your parameters. You can override this exception by setting input.ignore_gain_mismatch=True."""
                                % (panel.get_gain(), gain)
                            )

    def find_spots(self, experiments):
        st = time.time()

        logger.info("*" * 80)
        logger.info("Finding Strong Spots")
        logger.info("*" * 80)

        # Find the strong spots
        observed = flex.reflection_table.from_observations(
            experiments, self.params, is_stills=True
        )

        # Reset z coordinates for dials.image_viewer; see Issues #226 for details
        xyzobs = observed["xyzobs.px.value"]
        for i in range(len(xyzobs)):
            xyzobs[i] = (xyzobs[i][0], xyzobs[i][1], 0)
        bbox = observed["bbox"]
        for i in range(len(bbox)):
            bbox[i] = (bbox[i][0], bbox[i][1], bbox[i][2], bbox[i][3], 0, 1)

        if self.params.output.composite_output:
            n = len(self.all_strong_reflections.experiment_identifiers())
            for i, experiment in enumerate(experiments):
                refls = observed.select(observed["id"] == i)
                refls["id"] = flex.int(len(refls), n)
                del refls.experiment_identifiers()[i]
                refls.experiment_identifiers()[n] = experiment.identifier
                self.all_strong_reflections.extend(refls)
                n += 1
        else:
            # Save the reflections to file
            logger.info("\n" + "-" * 80)
            if self.params.output.strong_filename:
                self.save_reflections(observed, self.params.output.strong_filename)

        logger.info("")
        logger.info("Time Taken = %f seconds", time.time() - st)
        return observed

    def index(self, experiments, reflections):
        from dials.algorithms.indexing.indexer import Indexer

        st = time.time()

        logger.info("*" * 80)
        logger.info("Indexing Strong Spots")
        logger.info("*" * 80)

        params = copy.deepcopy(self.params)
        # don't do scan-varying refinement during indexing
        params.refinement.parameterisation.scan_varying = False

        if hasattr(self, "known_crystal_models"):
            known_crystal_models = self.known_crystal_models
        elif self.params.indexing.stills.known_orientations:
            known_crystal_models = []
            extended_experiments = ExperimentList()
            for expt in experiments:
                assert (
                    len(expt.imageset.indices()) == 1
                    and len(expt.imageset.paths()) == 1
                )
                key = (
                    os.path.basename(expt.imageset.paths()[0]),
                    expt.imageset.indices()[0],
                )
                if key not in self.params.indexing.stills.known_orientations:
                    if self.params.indexing.stills.require_known_orientation:
                        raise RuntimeError(
                            "Image not found in set of known orientations"
                        )
                    else:
                        oris = [None]
                else:
                    oris = self.params.indexing.stills.known_orientations[key]
                known_crystal_models.extend(oris)
                extended_experiments.extend(ExperimentList([expt] * len(oris)))
            experiments = extended_experiments
        else:
            known_crystal_models = None

        indexing_succeeded = False
        if known_crystal_models:
            try:
                idxr = Indexer.from_parameters(
                    reflections,
                    experiments,
                    known_crystal_models=known_crystal_models,
                    params=params,
                )
                idxr.index()
                logger.info("indexed from known orientation")
                indexing_succeeded = True
            except Exception:
                if self.params.indexing.stills.require_known_orientation:
                    raise

        if params.indexing.stills.method_list is None and not indexing_succeeded:
            idxr = Indexer.from_parameters(
                reflections,
                experiments,
                params=params,
            )
            idxr.index()
        elif not indexing_succeeded:
            indexing_error = None
            for method in params.indexing.stills.method_list:
                params.indexing.method = method
                try:
                    idxr = Indexer.from_parameters(
                        reflections, experiments, params=params
                    )
                    idxr.index()
                except Exception as e:
                    logger.info("Couldn't index using method %s", method)
                    if indexing_error is None:
                        if e is None:
                            e = Exception(f"Couldn't index using method {method}")
                        indexing_error = e
                else:
                    indexing_error = None
                    break
            if indexing_error is not None:
                raise indexing_error

        indexed = idxr.refined_reflections
        experiments = idxr.refined_experiments

        if known_crystal_models is not None:
            filtered_sel = flex.bool(len(indexed), True)
            for expt_id in range(len(experiments)):
                for idx in set(
                    indexed["miller_index"].select(indexed["id"] == expt_id)
                ):
                    sel = (indexed["miller_index"] == idx) & (indexed["id"] == expt_id)
                    if sel.count(True) > 1:
                        filtered_sel = filtered_sel & ~sel
            filtered = indexed.select(filtered_sel)
            logger.info(
                "Filtered duplicate reflections, %d out of %d remaining",
                len(filtered),
                len(indexed),
            )
            print(
                "Filtered duplicate reflections, %d out of %d remaining"
                % (len(filtered), len(indexed))
            )
            indexed = filtered

        logger.info("")
        logger.info("Time Taken = %f seconds", time.time() - st)
        return experiments, indexed

    def refine(self, experiments, centroids):
        if self.params.dispatch.refine:
            from dials.algorithms.refinement import RefinerFactory

            st = time.time()

            logger.info("*" * 80)
            logger.info("Refining Model")
            logger.info("*" * 80)

            refiner = RefinerFactory.from_parameters_data_experiments(
                self.params, centroids, experiments
            )

            refiner.run()
            experiments = refiner.get_experiments()
            predicted = refiner.predict_for_indexed()
            centroids["xyzcal.mm"] = predicted["xyzcal.mm"]
            centroids["entering"] = predicted["entering"]
            centroids = centroids.select(refiner.selection_used_for_refinement())

            # Re-estimate mosaic estimates
            from dials.algorithms.indexing.nave_parameters import NaveParameters

            nv = NaveParameters(
                params=self.params,
                experiments=experiments,
                reflections=centroids,
                refinery=refiner,
                graph_verbose=False,
            )
            nv()
            acceptance_flags_nv = nv.nv_acceptance_flags
            centroids = centroids.select(acceptance_flags_nv)

        if self.params.output.composite_output:
            if (
                self.params.output.refined_experiments_filename
                or self.params.output.indexed_filename
            ):
                assert (
                    self.params.output.refined_experiments_filename is not None
                    and self.params.output.indexed_filename is not None
                )

                n = len(self.all_indexed_experiments)
                self.all_indexed_experiments.extend(experiments)
                for i, experiment in enumerate(experiments):
                    refls = centroids.select(centroids["id"] == i)
                    refls["id"] = flex.int(len(refls), n)
                    del refls.experiment_identifiers()[i]
                    refls.experiment_identifiers()[n] = experiment.identifier
                    self.all_indexed_reflections.extend(refls)
                    n += 1
        else:
            # Dump experiments to disk
            if self.params.output.refined_experiments_filename:

                experiments.as_json(self.params.output.refined_experiments_filename)

            if self.params.output.indexed_filename:
                self.save_reflections(centroids, self.params.output.indexed_filename)

        if self.params.dispatch.refine:
            logger.info("")
            logger.info("Time Taken = %f seconds", time.time() - st)

        return experiments, centroids

    def integrate(self, experiments, indexed):
        st = time.time()

        logger.info("*" * 80)
        logger.info("Integrating Reflections")
        logger.info("*" * 80)

        indexed, _ = self.process_reference(indexed)

        if self.params.integration.integration_only_overrides.trusted_range:
            for detector in experiments.detectors():
                for panel in detector:
                    panel.set_trusted_range(
                        self.params.integration.integration_only_overrides.trusted_range
                    )

        if self.params.dispatch.coset:
            from xfel.util.sublattice_helper import integrate_coset

            integrate_coset(self, experiments, indexed)

        # Get the integrator from the input parameters
        logger.info("Configuring integrator from input parameters")
        from dials.algorithms.integration.integrator import create_integrator
        from dials.algorithms.profile_model.factory import ProfileModelFactory

        # Compute the profile model
        # Predict the reflections
        # Match the predictions with the reference
        # Create the integrator
        experiments = ProfileModelFactory.create(self.params, experiments, indexed)
        new_experiments = ExperimentList()
        new_reflections = flex.reflection_table()
        for expt_id, expt in enumerate(experiments):
            if (
                self.params.profile.gaussian_rs.parameters.sigma_b_cutoff is None
                or expt.profile.sigma_b()
                < self.params.profile.gaussian_rs.parameters.sigma_b_cutoff
            ):
                refls = indexed.select(indexed["id"] == expt_id)
                refls["id"] = flex.int(len(refls), len(new_experiments))
                # refls.reset_ids()
                del refls.experiment_identifiers()[expt_id]
                refls.experiment_identifiers()[len(new_experiments)] = expt.identifier
                new_reflections.extend(refls)
                new_experiments.append(expt)
            else:
                logger.info(
                    "Rejected expt %d with sigma_b %f"
                    % (expt_id, expt.profile.sigma_b())
                )
        experiments = new_experiments
        indexed = new_reflections
        if len(experiments) == 0:
            raise RuntimeError("No experiments after filtering by sigma_b")
        logger.info("")
        logger.info("=" * 80)
        logger.info("")
        logger.info("Predicting reflections")
        logger.info("")
        predicted = flex.reflection_table.from_predictions_multi(
            experiments,
            dmin=self.params.prediction.d_min,
            dmax=self.params.prediction.d_max,
            margin=self.params.prediction.margin,
            force_static=self.params.prediction.force_static,
        )
        predicted.match_with_reference(indexed)
        logger.info("")
        integrator = create_integrator(self.params, experiments, predicted)

        # Integrate the reflections
        integrated = integrator.integrate()

        # correct integrated intensities for absorption correction, if necessary
        for abs_params in self.params.integration.absorption_correction:
            if abs_params.apply:
                if abs_params.algorithm == "fuller_kapton":
                    from dials.algorithms.integration.kapton_correction import (
                        multi_kapton_correction,
                    )
                elif abs_params.algorithm == "kapton_2019":
                    from dials.algorithms.integration.kapton_2019_correction import (
                        multi_kapton_correction,
                    )

                experiments, integrated = multi_kapton_correction(
                    experiments, integrated, abs_params.fuller_kapton, logger=logger
                )()

        if self.params.significance_filter.enable:
            from dials.algorithms.integration.stills_significance_filter import (
                SignificanceFilter,
            )

            sig_filter = SignificanceFilter(self.params)
            filtered_refls = sig_filter(experiments, integrated)
            accepted_expts = ExperimentList()
            accepted_refls = flex.reflection_table()
            logger.info(
                "Removed %d reflections out of %d when applying significance filter",
                len(integrated) - len(filtered_refls),
                len(integrated),
            )
            for expt_id, expt in enumerate(experiments):
                refls = filtered_refls.select(filtered_refls["id"] == expt_id)
                if len(refls) > 0:
                    accepted_expts.append(expt)
                    refls["id"] = flex.int(len(refls), len(accepted_expts) - 1)
                    accepted_refls.extend(refls)
                else:
                    logger.info(
                        "Removed experiment %d which has no reflections left after applying significance filter",
                        expt_id,
                    )

            if len(accepted_refls) == 0:
                raise Sorry("No reflections left after applying significance filter")
            experiments = accepted_expts
            integrated = accepted_refls

        # Delete the shoeboxes used for intermediate calculations, if requested
        if self.params.integration.debug.delete_shoeboxes and "shoebox" in integrated:
            del integrated["shoebox"]

        if self.params.output.composite_output:
            if (
                self.params.output.integrated_experiments_filename
                or self.params.output.integrated_filename
            ):
                assert (
                    self.params.output.integrated_experiments_filename is not None
                    and self.params.output.integrated_filename is not None
                )

                n = len(self.all_integrated_experiments)
                self.all_integrated_experiments.extend(experiments)
                for i, experiment in enumerate(experiments):
                    refls = integrated.select(integrated["id"] == i)
                    refls["id"] = flex.int(len(refls), n)
                    del refls.experiment_identifiers()[i]
                    refls.experiment_identifiers()[n] = experiment.identifier
                    self.all_integrated_reflections.extend(refls)
                    n += 1
        else:
            # Dump experiments to disk
            if self.params.output.integrated_experiments_filename:

                experiments.as_json(self.params.output.integrated_experiments_filename)

            if self.params.output.integrated_filename:
                # Save the reflections
                self.save_reflections(
                    integrated, self.params.output.integrated_filename
                )

        self.write_integration_pickles(integrated, experiments)
        from dials.algorithms.indexing.stills_indexer import (
            calc_2D_rmsd_and_displacements,
        )

        rmsd_indexed, _ = calc_2D_rmsd_and_displacements(indexed)
        log_str = f"RMSD indexed (px): {rmsd_indexed:f}\n"
        for i in range(6):
            bright_integrated = integrated.select(
                (
                    integrated["intensity.sum.value"]
                    / flex.sqrt(integrated["intensity.sum.variance"])
                )
                >= i
            )
            if len(bright_integrated) > 0:
                rmsd_integrated, _ = calc_2D_rmsd_and_displacements(bright_integrated)
            else:
                rmsd_integrated = 0
            log_str += (
                "N reflections integrated at I/sigI >= %d: % 4d, RMSD (px): %f\n"
                % (i, len(bright_integrated), rmsd_integrated)
            )

        for crystal_model in experiments.crystals():
            if hasattr(crystal_model, "get_domain_size_ang"):
                log_str += ". Final ML model: domain size angstroms: {:f}, half mosaicity degrees: {:f}".format(
                    crystal_model.get_domain_size_ang(),
                    crystal_model.get_half_mosaicity_deg(),
                )

        logger.info(log_str)

        logger.info("")
        logger.info("Time Taken = %f seconds", time.time() - st)
        return integrated

    def write_integration_pickles(self, integrated, experiments, callback=None):
        """
        Write a serialized python dictionary with integrated intensities and other information
        suitible for use by cxi.merge or prime.postrefine.
        @param integrated Reflection table with integrated intensities
        @param experiments Experiment list. One integration pickle for each experiment will be created.
        @param callback Deriving classes can use callback to make further modifications to the dictionary
        before it is serialized. Callback should be a function with this signature:
        def functionname(params, outfile, frame), where params is the phil scope, outfile is the path
        to the pickle that will be saved, and frame is the python dictionary to be serialized.
        """
        if not hasattr(self.params.output, "integration_pickle"):
            return

        if self.params.output.integration_pickle is not None:
            from xfel.command_line.frame_extractor import ConstructFrame

            # Split everything into separate experiments for pickling
            for e_number, experiment in enumerate(experiments):
                e_selection = integrated["id"] == e_number
                reflections = integrated.select(e_selection)

                frame = ConstructFrame(reflections, experiment).make_frame()
                frame["pixel_size"] = experiment.detector[0].get_pixel_size()[0]

                if not hasattr(self, "tag") or self.tag is None:
                    try:
                        # if the data was a file on disc, get the path
                        event_timestamp = os.path.splitext(
                            experiments[0].imageset.paths()[0]
                        )[0]
                    except NotImplementedError:
                        # if the data is in memory only, check if the reader set a timestamp on the format object
                        event_timestamp = (
                            experiment.imageset.reader().get_format(0).timestamp
                        )
                    event_timestamp = os.path.basename(event_timestamp)
                    if event_timestamp.find("shot-") == 0:
                        event_timestamp = os.path.splitext(event_timestamp)[
                            0
                        ]  # micromanage the file name
                else:
                    event_timestamp = self.tag
                if hasattr(self.params.output, "output_dir"):
                    outfile = os.path.join(
                        self.params.output.output_dir,
                        self.params.output.integration_pickle
                        % (e_number, event_timestamp),
                    )
                else:
                    outfile = os.path.join(
                        os.path.dirname(self.params.output.integration_pickle),
                        self.params.output.integration_pickle
                        % (e_number, event_timestamp),
                    )

                if callback is not None:
                    callback(self.params, outfile, frame)

                if self.params.output.composite_output:
                    self.all_int_pickle_filenames.append(os.path.basename(outfile))
                    self.all_int_pickles.append(frame)
                else:
                    with open(outfile, "wb") as fh:
                        pickle.dump(frame, fh, protocol=pickle.HIGHEST_PROTOCOL)

    def process_reference(self, reference):
        """Load the reference spots."""
        if reference is None:
            return None, None
        st = time.time()
        assert "miller_index" in reference
        assert "id" in reference
        logger.info("Processing reference reflections")
        logger.info(" read %d strong spots", len(reference))
        mask = reference.get_flags(reference.flags.indexed)
        rubbish = reference.select(~mask)
        if mask.count(False) > 0:
            reference.del_selected(~mask)
            logger.info(" removing %d unindexed reflections", mask.count(True))
        if len(reference) == 0:
            raise Sorry(
                """
        Invalid input for reference reflections.
        Expected > %d indexed spots, got %d
      """
                % (0, len(reference))
            )
        mask = reference["miller_index"] == (0, 0, 0)
        if mask.count(True) > 0:
            rubbish.extend(reference.select(mask))
            reference.del_selected(mask)
            logger.info(" removing %d reflections with hkl (0,0,0)", mask.count(True))
        mask = reference["id"] < 0
        if mask.count(True) > 0:
            raise Sorry(
                """
        Invalid input for reference reflections.
        %d reference spots have an invalid experiment id
      """
                % mask.count(True)
            )
        logger.info(" using %d indexed reflections", len(reference))
        logger.info(" found %d junk reflections", len(rubbish))
        logger.info(" time taken: %g", time.time() - st)
        return reference, rubbish

    def save_reflections(self, reflections, filename):
        """Save the reflections to file."""
        st = time.time()
        logger.info("Saving %d reflections to %s", len(reflections), filename)
        reflections.as_file(filename)
        logger.info(" time taken: %g", time.time() - st)

    def finalize(self):
        """Perform any final operations"""
        if self.params.output.composite_output:
            if self.params.mp.composite_stride is not None:
                assert self.params.mp.method == "mpi"
                stride = self.params.mp.composite_stride

                from mpi4py import MPI

                comm = MPI.COMM_WORLD
                rank = comm.Get_rank()  # each process in MPI has a unique id, 0-indexed
                size = comm.Get_size()  # size: number of processes running in this job
                comm.barrier()

                if rank % stride == 0:
                    subranks = [rank + i for i in range(1, stride) if rank + i < size]
                    for i in range(len(subranks)):
                        logger.info("Rank %d waiting for sender", rank)
                        (
                            sender,
                            imported_experiments,
                            strong_reflections,
                            indexed_experiments,
                            indexed_reflections,
                            integrated_experiments,
                            integrated_reflections,
                            coset_experiments,
                            coset_reflections,
                            int_pickles,
                            int_pickle_filenames,
                        ) = comm.recv(source=MPI.ANY_SOURCE)
                        logger.info("Rank %d received data from rank %d", rank, sender)

                        def extend_with_bookkeeping(
                            src_expts, src_refls, dest_expts, dest_refls
                        ):
                            n = len(dest_refls.experiment_identifiers())
                            src_refls["id"] += n
                            idents = src_refls.experiment_identifiers()
                            keys = idents.keys()
                            values = idents.values()
                            for key in keys:
                                del idents[key]
                            for i, key in enumerate(keys):
                                idents[key + n] = values[i]
                            dest_expts.extend(src_expts)
                            dest_refls.extend(src_refls)

                        if len(imported_experiments) > 0:
                            extend_with_bookkeeping(
                                imported_experiments,
                                strong_reflections,
                                self.all_imported_experiments,
                                self.all_strong_reflections,
                            )

                        if len(indexed_experiments) > 0:
                            extend_with_bookkeeping(
                                indexed_experiments,
                                indexed_reflections,
                                self.all_indexed_experiments,
                                self.all_indexed_reflections,
                            )

                        if len(integrated_experiments) > 0:
                            extend_with_bookkeeping(
                                integrated_experiments,
                                integrated_reflections,
                                self.all_integrated_experiments,
                                self.all_integrated_reflections,
                            )

                        if len(coset_experiments) > 0:
                            extend_with_bookkeeping(
                                coset_experiments,
                                coset_reflections,
                                self.all_coset_experiments,
                                self.all_coset_reflections,
                            )

                        self.all_int_pickles.extend(int_pickles)
                        self.all_int_pickle_filenames.extend(int_pickle_filenames)

                else:
                    destrank = (rank // stride) * stride
                    logger.info(
                        "Rank %d sending results to rank %d",
                        rank,
                        (rank // stride) * stride,
                    )
                    comm.send(
                        (
                            rank,
                            self.all_imported_experiments,
                            self.all_strong_reflections,
                            self.all_indexed_experiments,
                            self.all_indexed_reflections,
                            self.all_integrated_experiments,
                            self.all_integrated_reflections,
                            self.all_coset_experiments,
                            self.all_coset_reflections,
                            self.all_int_pickles,
                            self.all_int_pickle_filenames,
                        ),
                        dest=destrank,
                    )

                    self.all_imported_experiments = (
                        self.all_strong_reflections
                    ) = (
                        self.all_indexed_experiments
                    ) = (
                        self.all_indexed_reflections
                    ) = (
                        self.all_integrated_experiments
                    ) = (
                        self.all_integrated_reflections
                    ) = (
                        self.all_coset_experiments
                    ) = (
                        self.all_coset_reflections
                    ) = self.all_int_pickles = self.all_integrated_reflections = []

            # Dump composite files to disk
            if (
                len(self.all_imported_experiments) > 0
                and self.params.output.experiments_filename
            ):

                self.all_imported_experiments.as_json(
                    self.params.output.experiments_filename
                )

            if (
                len(self.all_strong_reflections) > 0
                and self.params.output.strong_filename
            ):
                self.save_reflections(
                    self.all_strong_reflections, self.params.output.strong_filename
                )

            if (
                len(self.all_indexed_experiments) > 0
                and self.params.output.refined_experiments_filename
            ):

                self.all_indexed_experiments.as_json(
                    self.params.output.refined_experiments_filename
                )

            if (
                len(self.all_indexed_reflections) > 0
                and self.params.output.indexed_filename
            ):
                self.save_reflections(
                    self.all_indexed_reflections, self.params.output.indexed_filename
                )

            if (
                len(self.all_integrated_experiments) > 0
                and self.params.output.integrated_experiments_filename
            ):

                self.all_integrated_experiments.as_json(
                    self.params.output.integrated_experiments_filename
                )

            if (
                len(self.all_integrated_reflections) > 0
                and self.params.output.integrated_filename
            ):
                self.save_reflections(
                    self.all_integrated_reflections,
                    self.params.output.integrated_filename,
                )

            if self.params.dispatch.coset:
                if (
                    len(self.all_coset_experiments) > 0
                    and self.params.output.coset_experiments_filename
                ):

                    self.all_coset_experiments.as_json(
                        self.params.output.coset_experiments_filename
                    )

                if (
                    len(self.all_coset_reflections) > 0
                    and self.params.output.coset_filename
                ):
                    self.save_reflections(
                        self.all_coset_reflections, self.params.output.coset_filename
                    )

            # Create a tar archive of the integration dictionary pickles
            if len(self.all_int_pickles) > 0 and self.params.output.integration_pickle:
                tar_template_integration_pickle = (
                    self.params.output.integration_pickle.replace("%d", "%s")
                )
                outfile = (
                    os.path.join(
                        self.params.output.output_dir,
                        tar_template_integration_pickle % ("x", self.composite_tag),
                    )
                    + ".tar"
                )
                tar = tarfile.TarFile(outfile, "w")
                for i, (fname, d) in enumerate(
                    zip(self.all_int_pickle_filenames, self.all_int_pickles)
                ):
                    string = BytesIO(pickle.dumps(d, protocol=2))
                    info = tarfile.TarInfo(name=fname)
                    info.size = string.getbuffer().nbytes
                    info.mtime = time.time()
                    tar.addfile(tarinfo=info, fileobj=string)
                tar.close()


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



# LIBTBX_SET_DISPATCHER_NAME dev.dials.simple_integrate
from __future__ import annotations

import logging
from math import ceil

import cctbx.array_family.flex

import dials.util.log
from dials.algorithms.integration.report import IntegrationReport, ProfileModelReport
from dials.algorithms.profile_model.gaussian_rs import GaussianRSProfileModeller
from dials.algorithms.profile_model.gaussian_rs import Model as GaussianRSProfileModel
from dials.algorithms.profile_model.gaussian_rs.calculator import (
    ComputeEsdBeamDivergence,
    ComputeEsdReflectingRange,
)
from dials.algorithms.shoebox import MaskCode
from dials.array_family import flex
from dials.command_line.integrate import filter_reference_pixels, process_reference
from dials.extensions.simple_background_ext import SimpleBackgroundExt
from dials.extensions.simple_centroid_ext import SimpleCentroidExt
from dials.model.data import make_image
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.phil import parse
from dials.util.version import dials_version
from dials_algorithms_integration_integrator_ext import ShoeboxProcessor

logger = logging.getLogger("dials.command_line.simple_integrate")

phil_scope = parse(
    """
output {
reflections = 'simple_integrated.refl'
    .type = str
    .help = "The integrated output filename"

phil = 'dials.simple_integrate.phil'
    .type = str
    .help = "The output phil file"

log = 'dials.simple_integrate.log'
    .type = str
    .help = "The log filename"
}
"""
)

"""
Kabsch 2010 refers to
Kabsch W., Integration, scaling, space-group assignment and
post-refinment, Acta Crystallographica Section D, 2010, D66, 133-144

Usage:
$ dev.dials.simple_integrate.py refined.expt refined.refl
"""


def run():

    """
    Input setup
    """

    phil = phil_scope.fetch()

    usage = "usage: dev.dials.simple_integrate.py models.expt reflections.expt"
    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        epilog=__doc__,
        read_experiments=True,
        read_reflections=True,
    )

    params, options = parser.parse_args(args=None, show_diff_phil=False)

    dials.util.log.config(verbosity=options.verbose, logfile=params.output.log)
    logger.info(dials_version())

    """
    Load experiment and reflections
    """

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    reflections = reflections[0]

    reflections["id"] = cctbx.array_family.flex.int(len(reflections), 0)
    reflections["imageset_id"] = cctbx.array_family.flex.int(len(reflections), 0)

    integrated_reflections = run_simple_integrate(params, experiments, reflections)
    integrated_reflections.as_msgpack_file(params.output.reflections)


def run_simple_integrate(params, experiments, reflections):

    experiment = experiments[0]

    # Remove bad reflections (e.g. those not indexed)
    reflections, _ = process_reference(reflections)
    # Mask neighbouring pixels to shoeboxes
    reflections = filter_reference_pixels(reflections, experiments)

    """
    Predict reflections using experiment crystal
    """

    predicted_reflections = flex.reflection_table.from_predictions(
        experiment, padding=1.0
    )
    predicted_reflections["id"] = cctbx.array_family.flex.int(
        len(predicted_reflections), 0
    )
    predicted_reflections["imageset_id"] = cctbx.array_family.flex.int(
        len(predicted_reflections), 0
    )
    # Updates flags to set which reflections to use in generating reference profiles
    matched, reflections, unmatched = predicted_reflections.match_with_reference(
        reflections
    )

    """
    Create profile model and add it to experiment.
    This is used to predict reflection properties.
    """

    # Filter reflections to use to create the model
    min_zeta = 0.05
    used_in_ref = reflections.get_flags(reflections.flags.used_in_refinement)
    model_reflections = reflections.select(used_in_ref)
    zeta = model_reflections.compute_zeta(experiment)
    model_reflections = model_reflections.select(flex.abs(zeta) >= min_zeta)

    # sigma_D in 3.1 of Kabsch 2010
    sigma_b = ComputeEsdBeamDivergence(
        experiment.detector, model_reflections, centroid_definition="s1"
    ).sigma()

    # sigma_m in 3.1 of Kabsch 2010
    sigma_m = ComputeEsdReflectingRange(
        experiment.crystal,
        experiment.beam,
        experiment.detector,
        experiment.goniometer,
        experiment.scan,
        model_reflections,
        algorithm="extended",
    ).sigma()

    # The Gaussian model given in 2.3 of Kabsch 2010
    experiment.profile = GaussianRSProfileModel(
        params=params, n_sigma=3, sigma_b=sigma_b, sigma_m=sigma_m
    )

    """
    Compute properties for predicted reflections using profile model,
    accessed via experiment.profile_model. These reflection_table
    methods are largely just wrappers for profile_model.compute_bbox etc.

    Note: I do not think all these properties are needed for integration,
    but are all present in the current dials.integrate output.
    """

    predicted_reflections.compute_bbox(experiments)
    predicted_reflections.compute_d(experiments)
    min_zeta = 0.05
    zeta = predicted_reflections.compute_zeta(experiment)
    predicted_reflections = predicted_reflections.select(flex.abs(zeta) >= min_zeta)
    predicted_reflections.compute_partiality(experiments)

    # Shoeboxes
    predicted_reflections["shoebox"] = flex.shoebox(
        predicted_reflections["panel"],
        predicted_reflections["bbox"],
        allocate=False,
        flatten=False,
    )

    # Get actual shoebox values and the reflections for each image
    shoebox_processor = ShoeboxProcessor(
        predicted_reflections,
        len(experiment.detector),
        0,
        len(experiment.imageset),
        False,
    )

    for i in range(len(experiment.imageset)):
        image = experiment.imageset.get_corrected_data(i)
        mask = experiment.imageset.get_mask(i)
        shoebox_processor.next_data_only(make_image(image, mask))

    predicted_reflections.is_overloaded(experiments)
    predicted_reflections.compute_mask(experiments)
    predicted_reflections.contains_invalid_pixels()

    # Background calculated explicitly to expose underlying algorithm
    background_algorithm = SimpleBackgroundExt(params=None, experiments=experiments)
    success = background_algorithm.compute_background(predicted_reflections)
    predicted_reflections.set_flags(
        ~success, predicted_reflections.flags.failed_during_background_modelling
    )

    # Centroids calculated explicitly to expose underlying algorithm
    centroid_algorithm = SimpleCentroidExt(params=None, experiments=experiments)
    centroid_algorithm.compute_centroid(predicted_reflections)

    predicted_reflections.compute_summed_intensity()

    # Filter reflections with a high fraction of masked foreground
    valid_foreground_threshold = 0.75  # DIALS default
    sboxs = predicted_reflections["shoebox"]
    nvalfg = sboxs.count_mask_values(MaskCode.Valid | MaskCode.Foreground)
    nforeg = sboxs.count_mask_values(MaskCode.Foreground)
    fraction_valid = nvalfg.as_double() / nforeg.as_double()
    selection = fraction_valid < valid_foreground_threshold
    predicted_reflections.set_flags(
        selection, predicted_reflections.flags.dont_integrate
    )

    predicted_reflections["num_pixels.valid"] = sboxs.count_mask_values(MaskCode.Valid)
    predicted_reflections["num_pixels.background"] = sboxs.count_mask_values(
        MaskCode.Valid | MaskCode.Background
    )
    predicted_reflections["num_pixels.background_used"] = sboxs.count_mask_values(
        MaskCode.Valid | MaskCode.Background | MaskCode.BackgroundUsed
    )
    predicted_reflections["num_pixels.foreground"] = nvalfg

    """
    Load modeller that will calculate reference profiles and
    do the actual profile fitting integration.
    """

    # Default params when running dials.integrate with C2sum_1_*.cbf.gz
    fit_method = 1  # reciprocal space fitter (called explicitly below)
    grid_method = 2  # regular grid
    grid_size = 5  # Downsampling grid size described in 3.3 of Kabsch 2010
    # Get the number of scan points
    scan_step = 5
    scan_range = experiment.scan.get_oscillation_range(deg=True)
    scan_range = abs(scan_range[1] - scan_range[0])
    num_scan_points = int(ceil(scan_range / scan_step))
    n_sigma = 4.5  # multiplier to expand bounding boxes
    fitting_threshold = 0.02
    reference_profile_modeller = GaussianRSProfileModeller(
        experiment.beam,
        experiment.detector,
        experiment.goniometer,
        experiment.scan,
        sigma_b,
        sigma_m,
        n_sigma,
        grid_size,
        num_scan_points,
        fitting_threshold,
        grid_method,
        fit_method,
    )

    """
    Calculate grid of reference profiles from predicted reflections
    that matched observed.
    ("Learning phase" of 3.3 in Kabsch 2010)
    """

    sel = predicted_reflections.get_flags(predicted_reflections.flags.reference_spot)
    reference_reflections = predicted_reflections.select(sel)
    sel = reference_reflections.get_flags(reference_reflections.flags.dont_integrate)
    sel = ~sel
    reference_reflections = reference_reflections.select(sel)
    reference_profile_modeller.model(reference_reflections)
    reference_profile_modeller.normalize_profiles()

    profile_model_report = ProfileModelReport(
        experiments, [reference_profile_modeller], model_reflections
    )
    logger.info("")
    logger.info(profile_model_report.as_str(prefix=" "))

    """
    Carry out the integration by fitting to reference profiles in 1D.
    (Calculates intensity using 3.4 of Kabsch 2010)
    """

    sel = predicted_reflections.get_flags(predicted_reflections.flags.dont_integrate)
    sel = ~sel
    predicted_reflections = predicted_reflections.select(sel)
    reference_profile_modeller.fit_reciprocal_space(predicted_reflections)
    predicted_reflections.compute_corrections(experiments)

    integration_report = IntegrationReport(experiments, predicted_reflections)
    logger.info("")
    logger.info(integration_report.as_str(prefix=" "))

    """
    Filter for integrated reflections and remove shoeboxes
    """

    del predicted_reflections["shoebox"]
    sel = predicted_reflections.get_flags(
        predicted_reflections.flags.integrated, all=False
    )
    predicted_reflections = predicted_reflections.select(sel)
    return predicted_reflections


if __name__ == "__main__":
    run()



"""
Tests for scaling library module.
"""

from __future__ import annotations

import os
from unittest.mock import Mock

import pytest

from cctbx import uctbx
from cctbx.sgtbx import space_group
from dxtbx.model import Beam, Crystal, Detector, Experiment, Goniometer, Scan
from dxtbx.model.experiment_list import ExperimentList
from libtbx import phil

from dials.algorithms.scaling.model.model import KBScalingModel, PhysicalScalingModel
from dials.algorithms.scaling.scaling_library import (
    choose_initial_scaling_intensities,
    create_datastructures_for_structural_model,
    create_Ih_table,
    create_scaling_model,
    determine_best_unit_cell,
    scale_single_dataset,
    scaled_data_as_miller_array,
)
from dials.array_family import flex
from dials.util.options import ArgumentParser


@pytest.fixture
def test_reflections():
    """Make a test reflection table."""
    return generated_refl()


@pytest.fixture
def test_experiments():
    """Make a test experiments list"""
    return generated_exp()


@pytest.fixture()
def test_params():
    """Make a test param phil scope."""
    return generated_param()


@pytest.fixture()
def mock_cif():
    """Mock a cif file for testing loading data from a cif."""
    cif = Mock()
    cif.intensities = flex.double([1.0, 2.0])
    cif.indices = flex.miller_index([(1, 0, 0), (0, 0, 1)])
    cif.space_group = space_group("C 2y")
    return cif


def generated_refl():
    """Generate a reflection table."""
    # these miller_idx/d_values don't make physical sense, but I didn't want to
    # have to write the tests for lots of reflections.
    reflections = flex.reflection_table()
    reflections["intensity.prf.value"] = flex.double([1.0, 10.0, 10.0, 1.0, 2.0])
    reflections["intensity.prf.variance"] = flex.double([1.0, 10.0, 10.0, 1.0, 2.0])
    reflections["intensity.sum.value"] = flex.double([10.0, 100.0, 100.0, 10.0, 10.0])
    reflections["intensity.sum.variance"] = flex.double(
        [10.0, 100.0, 100.0, 10.0, 10.0]
    )
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (1, 0, 0), (0, 0, 1), (0, 0, 2)]
    )  # don't change
    reflections["d"] = flex.double([0.8, 2.0, 0.8, 2.0, 1.2])  # don't change
    reflections["lp"] = flex.double(5, 1.0)
    reflections["partiality"] = flex.double([1.0, 1.0, 1.0, 1.0, 0.8])
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 5.0),
            (0.0, 0.0, 10.0),
            (0.0, 0.0, 10.0),
            (0.0, 0.0, 7.5),
        ]
    )
    reflections["s1"] = flex.vec3_double(
        [
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
            (0.0, 0.1, 1.0),
        ]
    )
    reflections.set_flags(flex.bool(5, True), reflections.flags.integrated)
    reflections["id"] = flex.int(5, 0)
    reflections["partial_id"] = flex.int(range(0, 5))
    reflections.experiment_identifiers()[0] = str(0)
    return reflections


def generated_exp(n=1, scan=True, image_range=[0, 10]):
    """Generate an experiment list with two experiments."""
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    crystal = Crystal.from_dict(exp_dict)
    if scan:
        scan = Scan(image_range=image_range, oscillation=[0.0, 1.0])
    else:
        scan = None
    beam = Beam(s0=(0.0, 0.0, 1.01))
    goniometer = Goniometer((1.0, 0.0, 0.0))
    goniometer_2 = Goniometer((1.0, 1.0, 0.0))
    detector = Detector()
    experiments.append(
        Experiment(
            beam=beam,
            scan=scan,
            goniometer=goniometer,
            detector=detector,
            crystal=crystal,
        )
    )
    experiments[0].identifier = "0"
    if n > 1:
        for i in range(0, n - 1):
            experiments.append(
                Experiment(
                    beam=beam,
                    scan=scan,
                    goniometer=goniometer_2,
                    detector=detector,
                    crystal=crystal,
                )
            )
            experiments[i + 1].identifier = str(i + 1)
    return experiments


def generated_param(absorption_term=False):
    """Generate a param phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_scope
  """,
        process_includes=True,
    )
    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.physical.absorption_correction = absorption_term
    parameters.array.absorption_correction = absorption_term
    parameters.array.n_resolution_bins = 1  # to stop example dataset
    # being overparameterised for array model refinement.
    return parameters


def test_datastructure_from_structural_model(dials_data, test_experiments):
    pdb_file = os.fspath(dials_data("cunir_serial", pathlib=True) / "2BW4.pdb")
    expt, table = create_datastructures_for_structural_model(test_experiments, pdb_file)
    assert expt.scaling_model.is_scaled
    assert table["intensity"]
    assert table["miller_index"]
    table.assert_experiment_identifiers_are_consistent(ExperimentList([expt]))


@pytest.mark.parametrize("model", ["physical", "array"])
def test_scale_single_dataset(test_reflections, test_experiments, test_params, model):
    """Test completion of scaling."""
    scaled_reflections = scale_single_dataset(
        test_reflections, test_experiments, test_params, model=model
    )
    assert "inverse_scale_factor" in scaled_reflections
    assert "inverse_scale_factor_variance" in scaled_reflections
    # what about when no params supplied?


def test_scale_single_dataset_no_params_supplied(test_reflections, test_experiments):
    """Test when no params scope supplied."""
    scaled_reflections = scale_single_dataset(
        test_reflections, test_experiments, model="physical"
    )
    assert "inverse_scale_factor" in scaled_reflections
    assert "inverse_scale_factor_variance" in scaled_reflections


def test_create_scaling_model():
    """Test the create scaling model function."""

    # Test that one can create the correct scaling model with the phil param.
    for m in ["physical", "array", "KB"]:
        params = generated_param()
        exp = generated_exp()
        rt = generated_refl()
        params.model = m
        new_exp = create_scaling_model(params, exp, [rt])
        assert new_exp[0].scaling_model.id_ == m

    # If a scaling model already exists, then nothing else should happen.
    params = generated_param()
    exp = generated_exp()
    rt = generated_refl()
    exp[0].scaling_model = PhysicalScalingModel.from_data(params, exp[0], rt)
    old_scaling_model = exp[0].scaling_model
    params.model = "KB"
    new_exp = create_scaling_model(params, exp, [rt])
    new_scaling_model = new_exp[0].scaling_model
    assert new_scaling_model is old_scaling_model  # Should not modify original.

    # Test multiple datasets, where one already has a scaling model.
    exp = generated_exp(3)
    params = generated_param()
    rt = generated_refl()
    rt_2 = generated_refl()
    rt_3 = generated_refl()
    exp[0].scaling_model = PhysicalScalingModel.from_data(params, exp[0], rt)
    params.model = "KB"
    new_exp = create_scaling_model(params, exp, [rt, rt_2, rt_3])
    assert new_exp[0].scaling_model is exp[0].scaling_model
    assert isinstance(new_exp[1].scaling_model, KBScalingModel)
    assert isinstance(new_exp[2].scaling_model, KBScalingModel)

    # Now test overwrite_existing_models option
    params.overwrite_existing_models = True
    params.model = "physical"
    newer_exp = create_scaling_model(params, new_exp, [rt, rt_2, rt_3])
    for exp in newer_exp:
        assert isinstance(exp.scaling_model, PhysicalScalingModel)


def test_create_Ih_table(test_experiments, test_reflections):
    """Test the create_Ih_table function."""
    test_reflections["intensity"] = test_reflections["intensity.prf.value"]
    test_reflections["variance"] = test_reflections["intensity.prf.variance"]

    Ih_table = create_Ih_table(test_experiments, [test_reflections])

    # Test data has been sorted into one block as expected.
    assert list(Ih_table.blocked_data_list[0].asu_miller_index) == (
        [(0, 0, 1), (0, 0, 1), (0, 0, 2), (1, 0, 0), (1, 0, 0)]
    )

    selection = flex.bool([True, True, True, True, False])
    Ih_table = create_Ih_table(test_experiments, [test_reflections], [selection])
    assert list(Ih_table.blocked_data_list[0].asu_miller_index) == (
        [(0, 0, 1), (0, 0, 1), (1, 0, 0), (1, 0, 0)]
    )

    with pytest.raises(AssertionError):
        # Should fail if length of exp, refl and selection are different
        Ih_table = create_Ih_table(
            test_experiments, [test_reflections, test_reflections], [selection]
        )


def test_choose_initial_scaling_intensities(test_reflections):
    """Test for correct choice of intensities."""
    test_refl = test_reflections
    intstr = "prf"
    new_rt = choose_initial_scaling_intensities(test_refl, intstr)
    assert list(new_rt["intensity"]) == list(test_refl["intensity.prf.value"])
    assert list(new_rt["variance"]) == list(test_refl["intensity.prf.variance"])
    intstr = "sum"  # should apply partiality correction
    new_rt = choose_initial_scaling_intensities(test_refl, intstr)
    assert list(new_rt["intensity"]) == list(
        test_refl["intensity.sum.value"] / test_refl["partiality"]
    )
    assert list(new_rt["variance"]) == pytest.approx(
        list(test_refl["intensity.sum.variance"] / flex.pow2(test_refl["partiality"]))
    )


def test_auto_scaling_model():
    """Test auto options for scaling model creation."""
    params = generated_param()
    exp = generated_exp(scan=False)
    rt = generated_refl()
    params.model = "auto"
    new_exp = create_scaling_model(params, exp, [rt])
    assert new_exp[0].scaling_model.id_ == "KB"

    params = generated_param(absorption_term="auto")
    exp = generated_exp(image_range=[1, 5])  # 5 degree wedge
    params.model = "auto"
    new_exp = create_scaling_model(params, exp, [rt])
    assert new_exp[0].scaling_model.id_ == "physical"
    assert len(new_exp[0].scaling_model.components["scale"].parameters) == 5
    assert len(new_exp[0].scaling_model.components["decay"].parameters) == 3
    assert "absorption" not in new_exp[0].scaling_model.components

    params = generated_param(absorption_term="auto")
    exp = generated_exp(image_range=[1, 20])  # 20 degree wedge
    params.model = "auto"
    new_exp = create_scaling_model(params, exp, [rt])
    assert new_exp[0].scaling_model.id_ == "physical"
    assert len(new_exp[0].scaling_model.components["scale"].parameters) == 7
    assert len(new_exp[0].scaling_model.components["decay"].parameters) == 6
    assert "absorption" not in new_exp[0].scaling_model.components

    params = generated_param(absorption_term="auto")
    exp = generated_exp(image_range=[1, 75])  # 20 degree wedge
    params.model = "auto"
    new_exp = create_scaling_model(params, exp, [rt])
    assert new_exp[0].scaling_model.id_ == "physical"
    assert len(new_exp[0].scaling_model.components["scale"].parameters) == 12
    assert len(new_exp[0].scaling_model.components["decay"].parameters) == 10
    assert "absorption" in new_exp[0].scaling_model.components

    # Now test overwrite_existing_models option
    params.overwrite_existing_models = True
    params.model = "KB"
    newer_exp = create_scaling_model(params, new_exp, [rt])
    assert isinstance(newer_exp[0].scaling_model, KBScalingModel)


def test_determine_best_unit_cell(test_experiments):
    assert determine_best_unit_cell(test_experiments).parameters() == pytest.approx(
        test_experiments[0].crystal.get_unit_cell().parameters()
    )

    recalc_uc = uctbx.unit_cell((1.1, 1.1, 2.1, 90, 90, 90))
    test_experiments[0].crystal.set_recalculated_unit_cell(recalc_uc)
    assert determine_best_unit_cell(test_experiments).parameters() == pytest.approx(
        recalc_uc.parameters()
    )


def test_scaled_data_as_miller_array(dials_data):
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    reflections = flex.reflection_table.from_file(location / "scaled_20_25.refl")
    experiments = ExperimentList.from_file(
        location / "scaled_20_25.expt", check_format=False
    )
    reflections = [reflections.select(reflections["id"] == i) for i in range(2)]
    print([refl.size() for refl in reflections])

    miller_array = scaled_data_as_miller_array(reflections, experiments)
    assert miller_array.size() == 5503
    assert miller_array.info().wavelength == experiments[0].beam.get_wavelength()

    miller_array = scaled_data_as_miller_array(reflections, experiments, wavelength=1)
    assert miller_array.size() == 5503
    assert miller_array.info().wavelength == 1



"""Simulate a rotation dataset with a smoothly-varying beam position for
refinement testing. Script based on tst_nanoBragg_basic.py"""


from __future__ import annotations

import math

from cctbx.eltbx import henke
from dxtbx.model import (
    BeamFactory,
    Crystal,
    DetectorFactory,
    GoniometerFactory,
    ScanFactory,
)
from iotbx import pdb
from scitbx import matrix

_pdb_lines = """HEADER TEST
CRYST1   50.000   60.000   70.000  90.00  90.00  90.00 P 1
ATOM      1  O   HOH A   1      56.829   2.920  55.702  1.00 20.00           O
ATOM      2  O   HOH A   2      49.515  35.149  37.665  1.00 20.00           O
ATOM      3  O   HOH A   3      52.667  17.794  69.925  1.00 20.00           O
ATOM      4  O   HOH A   4      40.986  20.409  18.309  1.00 20.00           O
ATOM      5  O   HOH A   5      46.896  37.790  41.629  1.00 20.00           O
ATOM      6 SED  MSE A   6       1.000   2.000   3.000  1.00 20.00          SE
END
"""


class Simulation:
    def __init__(self, override_fdp=None):

        # Set up detector
        distance = 100
        pixel_size = 0.1
        image_size = (1000, 1000)
        beam_centre_mm = (
            pixel_size * image_size[0] / 2,
            pixel_size * image_size[1] / 2,
        )
        self.detector = DetectorFactory().simple(
            "CCD",
            distance,
            beam_centre_mm,
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
        )

        # Set up beam
        self.beam = BeamFactory().simple(wavelength=1)

        # Set up scan
        sequence_width = 90.0
        osc_start = 0.0
        image_width = 0.2
        oscillation = (osc_start, image_width)

        nframes = int(math.ceil(sequence_width / image_width))
        image_range = (1, nframes)
        exposure_times = 0.0
        epochs = [0] * nframes
        self.scan = ScanFactory().make_scan(
            image_range, exposure_times, oscillation, epochs, deg=True
        )

        # Set up goniometer
        self.goniometer = GoniometerFactory.known_axis(self.detector[0].get_fast_axis())

        # Set up simulated structure factors
        self.sfall = self.fcalc_from_pdb(
            resolution=1.6, algorithm="direct", override_fdp=override_fdp
        )

        # Set up crystal
        self.crystal = Crystal(
            real_space_a=(50, 0, 0),
            real_space_b=(0, 60, 0),
            real_space_c=(0, 0, 70),
            space_group_symbol="P1",
        )
        axis = matrix.col(
            elems=(-0.14480368275412925, -0.6202131724405818, -0.7709523423610766)
        )
        self.crystal.set_U(
            axis.axis_and_angle_as_r3_rotation_matrix(angle=0.625126343998969)
        )

    def fcalc_from_pdb(self, resolution, algorithm=None, override_fdp=None):
        pdb_inp = pdb.input(source_info=None, lines=_pdb_lines)
        xray_structure = pdb_inp.xray_structure_simple()
        wavelength = self.beam.get_wavelength()
        #
        # take a detour to calculate anomalous contribution of every atom
        scatterers = xray_structure.scatterers()
        for sc in scatterers:
            expected_henke = henke.table(sc.element_symbol()).at_angstrom(wavelength)
            sc.fp = expected_henke.fp()
            sc.fdp = override_fdp if override_fdp is not None else expected_henke.fdp()

        # how do we do bulk solvent?
        primitive_xray_structure = xray_structure.primitive_setting()
        P1_primitive_xray_structure = primitive_xray_structure.expand_to_p1()
        fcalc = P1_primitive_xray_structure.structure_factors(
            d_min=resolution, anomalous_flag=True, algorithm=algorithm
        ).f_calc()
        return fcalc.amplitudes()

    def set_varying_beam(self, along="fast", npixels_drift=5):
        assert along in ["fast", "slow", "both"]
        num_scan_points = self.scan.get_num_images() + 1
        s0 = matrix.col(self.beam.get_s0())
        beam_centre_px = self.detector[0].get_beam_centre_px(s0)
        if along == "fast":
            start_beam_centre = (
                beam_centre_px[0] - npixels_drift / 2,
                beam_centre_px[1],
            )
            end_beam_centre = (beam_centre_px[0] + npixels_drift / 2, beam_centre_px[1])
        elif along == "slow":
            start_beam_centre = (
                beam_centre_px[0],
                beam_centre_px[1] - npixels_drift / 2,
            )
            end_beam_centre = (beam_centre_px[0], beam_centre_px[1] + npixels_drift / 2)
        elif along == "both":
            offset = math.sqrt(2.0) * npixels_drift / 4.0
            start_beam_centre = (beam_centre_px[0] - offset, beam_centre_px[1] - offset)
            end_beam_centre = (beam_centre_px[0] + offset, beam_centre_px[1] + offset)

        start_lab = matrix.col(self.detector[0].get_pixel_lab_coord(start_beam_centre))
        end_lab = matrix.col(self.detector[0].get_pixel_lab_coord(end_beam_centre))
        axis = start_lab.cross(end_lab).normalize()
        full_angle = start_lab.angle(end_lab)
        angle_step = full_angle / self.scan.get_num_images()
        angles = [e * angle_step for e in range(num_scan_points)]

        start_s0 = start_lab.normalize() * s0.length()
        s0_list = [start_s0.rotate_around_origin(axis=axis, angle=e) for e in angles]

        self.beam.set_s0_at_scan_points(s0_list)



from __future__ import annotations

import pytest


def test(dials_data):
    from dxtbx.model.experiment_list import ExperimentListFactory

    from dials.algorithms.spot_prediction import PixelToMillerIndex
    from dials.array_family import flex

    filename = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    experiments = ExperimentListFactory.from_json_file(filename)

    transform = PixelToMillerIndex(
        experiments[0].beam,
        experiments[0].detector,
        experiments[0].goniometer,
        experiments[0].scan,
        experiments[0].crystal,
    )

    reflections = flex.reflection_table.from_predictions(experiments[0])

    for r in reflections.rows():
        panel = r["panel"]
        x, y, z = r["xyzcal.px"]
        h0 = r["miller_index"]
        h1 = transform.h(panel, x, y, z)
        assert h0 == pytest.approx(h1, abs=1e-7)



from __future__ import annotations

import pytest

from dials.command_line.frame_orientations import extract_experiment_data


def test_extract_experiment_data():
    """Test basic operation of the extract_experiment_data function. Does not
    test extraction of data from scan-varying models"""

    # Set up an Experiment with idealised geometry
    from dxtbx.model import BeamFactory, Crystal, GoniometerFactory, ScanFactory
    from dxtbx.model.experiment_list import Experiment

    beam = BeamFactory.make_beam(unit_s0=(0, 0, -1), wavelength=1.0)
    goniometer = GoniometerFactory.known_axis((1, 0, 0))
    a = (100, 0, 0)
    b = (0, 90, 0)
    c = (0, 0, 80)
    crystal = Crystal(a, b, c, space_group_symbol="P1")
    scan = ScanFactory.make_scan(
        image_range=(1, 91),
        exposure_times=0.1,
        oscillation=(-0.5, 1.0),
        epochs=list(range(91)),
        deg=True,
    )

    exp = Experiment(beam=beam, goniometer=goniometer, scan=scan, crystal=crystal)

    # Extract experiment data
    dat = extract_experiment_data(exp, scale=100)

    # Check results are as expected
    za = dat["zone_axes"]

    # At the middle of the first image the c axis is aligned antiparallel with
    # the beam vector, while the a and b axes are orthogonal. The zone axis
    # calculation is scaled by 100 (i.e. the max cell dimension, which is the
    # default), therefore we expect the zone axis [uvw] = [0 0 -100/80]
    assert za[0].elems == pytest.approx((0, 0, -100 / 80))

    # At the middle of the 91st image the crystal has rotated by 90 degrees, so
    # now c is orthogonal to the beam while b is anti-parallel to it. The zone
    # axis is now expected to be [uvw] = [0 -100/90 0]
    assert za[-1].elems == pytest.approx((0, -100 / 90, 0))

    rsa = dat["real_space_axes"]
    a, b, c = rsa[0]
    assert a.elems == pytest.approx(crystal.get_real_space_vectors()[0])
    assert b.elems == pytest.approx(crystal.get_real_space_vectors()[1])
    assert c.elems == pytest.approx(crystal.get_real_space_vectors()[2])



from __future__ import annotations

import logging
import math

import pkg_resources

import iotbx.phil
import libtbx
from cctbx import sgtbx
from dxtbx.model import ExperimentList, ImageSequence

import dials.util
from dials.algorithms.indexing import (
    DialsIndexError,
    DialsIndexRefineError,
    assign_indices,
)
from dials.algorithms.indexing.compare_orientation_matrices import (
    difference_rotation_matrix_axis_angle,
)
from dials.algorithms.indexing.max_cell import find_max_cell
from dials.algorithms.indexing.symmetry import SymmetryHandler
from dials.algorithms.refinement import DialsRefineConfigError, DialsRefineRuntimeError
from dials.array_family import flex
from dials.util.multi_dataset_handling import generate_experiment_identifiers

logger = logging.getLogger(__name__)


max_cell_phil_str = """\
max_cell_estimation
  .expert_level = 1
{
  filter_ice = True
    .type = bool
    .help = "Filter out reflections at typical ice ring resolutions"
            "before max_cell estimation."
  filter_overlaps = True
    .type = bool
    .help = "Filter out reflections with overlapping bounding boxes before"
            "max_cell estimation."
  overlaps_border = 0
    .type = int(value_min=0)
    .help = "Optionally add a border around the bounding boxes before finding"
            "overlaps."
  multiplier = 1.3
    .type = float(value_min=0)
    .help = "Multiply the estimated maximum basis vector length by this value."
    .expert_level = 2
  step_size = 45
    .type = float(value_min=0)
    .help = "Step size, in degrees, of the blocks used to perform the max_cell "
            "estimation."
    .expert_level = 2
  nearest_neighbor_percentile = None
    .type = float(value_min=0, value_max=1)
    .help = "Percentile of NN histogram to use for max cell determination."
    .expert_level = 2
  histogram_binning = linear *log
    .type = choice
    .help = "Choose between linear or logarithmic bins for nearest neighbour"
            "histogram analysis."
  nn_per_bin = 5
    .type = int(value_min=1)
    .help = "Target number of nearest neighbours per histogram bin."
  max_height_fraction = 0.25
    .type = float(value_min=0, value_max=1)
    .expert_level=2
}
"""

phil_str = (
    """\
indexing {
  nproc = 1
    .type = int(value_min=1)
    .help = "The number of processes to use."
  mm_search_scope = 4.0
    .help = "Global radius of origin offset search."
    .type = float(value_min=0)
    .expert_level = 1
  wide_search_binning = 2
    .help = "Modify the coarseness of the wide grid search for the beam centre."
    .type = float(value_min=0)
    .expert_level = 1
  min_cell_volume = 25
    .type = float(value_min=0)
    .help = "Minimum unit cell volume (in Angstrom^3)."
    .expert_level = 1
  min_cell = 3
    .type = float(value_min=0)
    .help = "Minimum length of candidate unit cell basis vectors (in Angstrom)."
    .expert_level = 1
  max_cell = Auto
    .type = float(value_min=0)
    .help = "Maximum length of candidate unit cell basis vectors (in Angstrom)."
    .expert_level = 1
  %s
  sigma_phi_deg = None
    .type = float(value_min=0)
    .help = "Override the phi sigmas for refinement. Mainly intended for single-shot"
            "rotation images where the phi sigma is almost certainly incorrect."
    .expert_level = 2

  known_symmetry {
    space_group = None
      .type = space_group
      .help = "Target space group for indexing."
    unit_cell = None
      .type = unit_cell
      .help = "Target unit cell for indexing."
    relative_length_tolerance = 0.1
      .type = float
      .help = "Relative tolerance for unit cell lengths in unit cell comparison."
      .expert_level = 1
    absolute_angle_tolerance = 5
      .type = float
      .help = "Angular tolerance (in degrees) in unit cell comparison."
      .expert_level = 1
    max_delta = 5
      .type = float(value_min=0)
      .help = "Maximum allowed Le Page delta used in searching for basis vector"
              "combinations that are consistent with the given symmetry."
      .expert_level = 1
  }

  index_assignment {
    method = *simple local
      .type = choice
      .help = "Choose between simple 'global' index assignment and xds-style "
              "'local' index assignment."
      .expert_level = 1
    simple {
      hkl_tolerance = 0.3
        .type = float(value_min=0, value_max=0.5)
        .help = "Maximum allowable deviation from integer-ness for assigning "
                "a miller index to a reciprocal lattice vector."
    }
    local
      .expert_level = 1
    {
      epsilon = 0.05
        .type = float
        .help = "This corresponds to the xds parameter INDEX_ERROR="
      delta = 8
        .type = int
        .help = "This corresponds to the xds parameter INDEX_MAGNITUDE="
      l_min = 0.8
        .type = float
        .help = "This corresponds to the xds parameter INDEX_QUALITY="
      nearest_neighbours = 20
        .type = int(value_min=1)
    }
  }
  check_misindexing {
    grid_search_scope = 0
      .type = int
      .help = "Search scope for testing misindexing on h, k, l."
  }
  debug = False
    .type = bool
    .expert_level = 1
  combine_scans = False
    .type = bool
    .expert_level = 1
  refinement_protocol {
    mode = *refine_shells repredict_only None
      .type = choice
      .expert_level = 1
      .help = "refine_shells: refine in increasing resolution cutoffs after indexing."
              "repredict_only: do not refine after indexing, just update spot"
              "predictions."
              "None: turns off all forms of refinement (currently only applies to stills_indexer)"
    n_macro_cycles = 5
      .type = int(value_min=1)
      .help = "Maximum number of macro cycles of refinement, reindexing all"
              "reflections using updated geometry at the beginning of each"
              "cycle. Does not apply to stills.indexer=stills."
    d_min_step = Auto
      .type = float(value_min=0.0)
      .help = "Reduction per step in d_min for reflections to include in refinement."
    d_min_start = None
      .type = float(value_min=0.0)
    d_min_final = None
      .type = float(value_min=0.0)
      .help = "Do not ever include reflections below this value in refinement."
    disable_unit_cell_volume_sanity_check = False
      .type = bool
      .help = "Disable sanity check on unrealistic increases in unit cell volume"
              "during refinement."
      .expert_level = 1
  }
  multiple_lattice_search
    .expert_level = 1
  {
    recycle_unindexed_reflections_cutoff = 0.1
      .type = float(value_min=0, value_max=1)
      .help = "Attempt another cycle of indexing on the unindexed reflections "
              "if more than the fraction of input reflections are unindexed."
    minimum_angular_separation = 5
      .type = float(value_min=0)
      .help = "The minimum angular separation (in degrees) between two lattices."
    max_lattices = 1
      .type = int
    cluster_analysis {
      method = *dbscan hcluster
        .type = choice
      hcluster {
        linkage {
          method = *ward
            .type = choice
          metric = *euclidean
            .type = choice
        }
        cutoff = 15
          .type = float(value_min=0)
        cutoff_criterion = *distance inconsistent
          .type = choice
      }
      dbscan {
        eps = 0.05
          .type = float(value_min=0.0)
        min_samples = 30
          .type = int(value_min=1)
      }
      min_cluster_size = 20
        .type = int(value_min=0)
      intersection_union_ratio_cutoff = 0.4
        .type = float(value_min=0.0, value_max=1.0)
    }
  }
  stills {
    indexer = *Auto stills sequences
      .type = choice
      .help = Use the stills or sequences indexer.  Auto: choose based on the input \
              imagesets (stills or sequences).
      .expert_level = 1
    ewald_proximity_resolution_cutoff = 2.0
      .type = float
      .help = For still images, this high-resolution cutoff is used to calculate
      .help = the acceptable volume of reciprocal space for spot prediction
    refine_all_candidates = True
      .type = bool
      .help = If False, no attempt is made to refine the model from initial basis \
              vector selection. The indexing solution with the best RMSD is chosen.
    candidate_outlier_rejection = True
      .type = bool
      .expert_level = 1
      .help = If True, while refining candidate basis solutions, also apply Sauter/ \
              Poon (2010) outlier rejection
    refine_candidates_with_known_symmetry = False
      .type = bool
      .expert_level = 2
      .help = If False, when choosing the best set of candidate basis solutions, \
              refine the candidates in the P1 setting. If True, after indexing \
              in P1, convert the candidates to the known symmetry and apply the \
              corresponding change of basis to the indexed reflections.
    rmsd_min_px = 2
      .type = float
      .help = Minimum acceptable RMSD for choosing candidate basis solutions \
              (in pixels)
    ewald_proximal_volume_max = 0.0025
      .type = float
      .help = Maximum acceptable ewald proximal volume when choosing candidate \
              basis solutions
    isoforms
      .help = Constrain the unit cell to specific values during refinement after initial indexing.
      .multiple=True
    {
      name=None
        .type=str
      cell=None
        .type=unit_cell
      lookup_symbol=None
        .type=str
        .help=The sgtbx lookup symbol of the reflections pointgroup
      rmsd_target_mm=None
        .type=float
        .help=Maximum acceptable DIALS positional rmsd, in mm
      beam_restraint=None
        .type=floats(size=2)
        .help=Known beam position in mm X,Y, rmsd_target_mm is reused here as a circle of confusion
        .help=to assure that no images are accepted where the lattice is misindexed by a unit shift.
    }
    set_domain_size_ang_value = None
      .type=float
      .help=If specified, will set the domain size ang value and override the value determined from nave refinement
    set_mosaic_half_deg_value = None
      .type=float
      .help=If specified, will set the mosaic half degree value and override the value determined from nave refinement
  }
}
"""
    % max_cell_phil_str
)

phil_scope = iotbx.phil.parse(phil_str, process_includes=True)


class Indexer:
    def __init__(self, reflections, experiments, params):
        self.reflections = reflections
        self.experiments = experiments

        self.params = params.indexing
        self.all_params = params
        self.refined_experiments = None
        self.hkl_offset = None

        if self.params.index_assignment.method == "local":
            self._assign_indices = assign_indices.AssignIndicesLocal(
                epsilon=self.params.index_assignment.local.epsilon,
                delta=self.params.index_assignment.local.delta,
                l_min=self.params.index_assignment.local.l_min,
                nearest_neighbours=self.params.index_assignment.local.nearest_neighbours,
            )
        else:
            self._assign_indices = assign_indices.AssignIndicesGlobal(
                tolerance=self.params.index_assignment.simple.hkl_tolerance
            )

        if self.all_params.refinement.reflections.outlier.algorithm in (
            "auto",
            libtbx.Auto,
        ):
            if self.experiments[0].goniometer is None:
                self.all_params.refinement.reflections.outlier.algorithm = "sauter_poon"
            else:
                # different default to dials.refine
                # tukey is faster and more appropriate at the indexing step
                self.all_params.refinement.reflections.outlier.algorithm = "tukey"

        for expt in self.experiments[1:]:
            if expt.detector.is_similar_to(self.experiments[0].detector):
                expt.detector = self.experiments[0].detector
            if expt.goniometer is not None and expt.goniometer.is_similar_to(
                self.experiments[0].goniometer
            ):
                expt.goniometer = self.experiments[0].goniometer
                # can only share a beam if we share a goniometer?
                if expt.beam.is_similar_to(self.experiments[0].beam):
                    expt.beam = self.experiments[0].beam
                if self.params.combine_scans and expt.scan == self.experiments[0].scan:
                    expt.scan = self.experiments[0].scan

        if "flags" in self.reflections:
            strong_sel = self.reflections.get_flags(self.reflections.flags.strong)
            if strong_sel.count(True) > 0:
                self.reflections = self.reflections.select(strong_sel)
        if "flags" not in self.reflections or strong_sel.count(True) == 0:
            # backwards compatibility for testing
            self.reflections.set_flags(
                flex.size_t_range(len(self.reflections)), self.reflections.flags.strong
            )

        self._setup_symmetry()
        self.d_min = None

        self.setup_indexing()

    @staticmethod
    def from_parameters(
        reflections, experiments, known_crystal_models=None, params=None
    ):

        if known_crystal_models is not None:
            from dials.algorithms.indexing.known_orientation import (
                IndexerKnownOrientation,
            )

            if params.indexing.known_symmetry.space_group is None:
                params.indexing.known_symmetry.space_group = (
                    known_crystal_models[0].get_space_group().info()
                )
            idxr = IndexerKnownOrientation(
                reflections, experiments, params, known_crystal_models
            )
        else:
            has_stills = False
            has_sequences = False
            for expt in experiments:
                if isinstance(expt.imageset, ImageSequence):
                    has_sequences = True
                else:
                    has_stills = True

            if has_stills and has_sequences:
                raise ValueError(
                    "Please provide only stills or only sequences, not both"
                )

            use_stills_indexer = has_stills

            if not (
                params.indexing.stills.indexer is libtbx.Auto
                or params.indexing.stills.indexer.lower() == "auto"
            ):
                if params.indexing.stills.indexer == "stills":
                    use_stills_indexer = True
                elif params.indexing.stills.indexer == "sequences":
                    use_stills_indexer = False
                else:
                    assert False

            if params.indexing.basis_vector_combinations.max_refine is libtbx.Auto:
                if use_stills_indexer:
                    params.indexing.basis_vector_combinations.max_refine = 5
                else:
                    params.indexing.basis_vector_combinations.max_refine = 50

            if use_stills_indexer:
                # Ensure the indexer and downstream applications treat this as set of stills
                from dxtbx.imageset import ImageSet  # , MemImageSet

                for experiment in experiments:
                    # Elsewhere, checks are made for ImageSequence when picking between algorithms
                    # specific to rotations vs. stills, so here reset any ImageSequences to stills.
                    # Note, dials.stills_process resets ImageSequences to ImageSets already,
                    # and it's not free (the ImageSet cache is dropped), only do it if needed
                    if isinstance(experiment.imageset, ImageSequence):
                        experiment.imageset = ImageSet(
                            experiment.imageset.data(), experiment.imageset.indices()
                        )
                    # if isinstance(imageset, MemImageSet):
                    #   imageset = MemImageSet(imagesequence._images, imagesequence.indices())
                    # else:
                    #   imageset = ImageSet(imagesequence.reader(), imagesequence.indices())
                    #   imageset._models = imagesequence._models
                    experiment.imageset.set_scan(None)
                    experiment.imageset.set_goniometer(None)
                    experiment.scan = None
                    experiment.goniometer = None

            IndexerType = None
            for entry_point in pkg_resources.iter_entry_points(
                "dials.index.basis_vector_search"
            ):
                if params.indexing.method == entry_point.name:
                    if use_stills_indexer:
                        # do something
                        from dials.algorithms.indexing.stills_indexer import (
                            StillsIndexerBasisVectorSearch as IndexerType,
                        )
                    else:
                        from dials.algorithms.indexing.lattice_search import (
                            BasisVectorSearch as IndexerType,
                        )

            if IndexerType is None:
                for entry_point in pkg_resources.iter_entry_points(
                    "dials.index.lattice_search"
                ):
                    if params.indexing.method == entry_point.name:
                        if use_stills_indexer:
                            from dials.algorithms.indexing.stills_indexer import (
                                StillsIndexerLatticeSearch as IndexerType,
                            )
                        else:
                            from dials.algorithms.indexing.lattice_search import (
                                LatticeSearch as IndexerType,
                            )

            assert IndexerType is not None

            idxr = IndexerType(reflections, experiments, params=params)

        return idxr

    def _setup_symmetry(self):
        target_unit_cell = self.params.known_symmetry.unit_cell
        target_space_group = self.params.known_symmetry.space_group
        if target_space_group is not None:
            target_space_group = target_space_group.group()
        else:
            target_space_group = sgtbx.space_group()
            self.params.known_symmetry.space_group = target_space_group.info()
        self._symmetry_handler = SymmetryHandler(
            unit_cell=target_unit_cell,
            space_group=target_space_group,
            max_delta=self.params.known_symmetry.max_delta,
        )
        return

    def setup_indexing(self):
        if len(self.reflections) == 0:
            raise DialsIndexError("No reflections left to index!")

        if "imageset_id" not in self.reflections:
            self.reflections["imageset_id"] = self.reflections["id"]
        self.reflections.centroid_px_to_mm(self.experiments)
        self.reflections.map_centroids_to_reciprocal_space(self.experiments)
        self.reflections.calculate_entering_flags(self.experiments)

        self.find_max_cell()

        if self.params.sigma_phi_deg is not None:
            var_x, var_y, _ = self.reflections["xyzobs.mm.variance"].parts()
            var_phi_rad = flex.double(
                var_x.size(), (math.pi / 180 * self.params.sigma_phi_deg) ** 2
            )
            self.reflections["xyzobs.mm.variance"] = flex.vec3_double(
                var_x, var_y, var_phi_rad
            )

        if self.params.debug:
            self._debug_write_reciprocal_lattice_points_as_pdb()

        self.reflections["id"] = flex.int(len(self.reflections), -1)

    def index(self):
        experiments = ExperimentList()

        had_refinement_error = False
        have_similar_crystal_models = False

        while True:
            if had_refinement_error or have_similar_crystal_models:
                break
            max_lattices = self.params.multiple_lattice_search.max_lattices
            if max_lattices is not None and len(experiments) >= max_lattices:
                break
            if len(experiments) > 0:
                cutoff_fraction = (
                    self.params.multiple_lattice_search.recycle_unindexed_reflections_cutoff
                )
                d_spacings = 1 / self.reflections["rlp"].norms()
                d_min_indexed = flex.min(d_spacings.select(self.indexed_reflections))
                min_reflections_for_indexing = cutoff_fraction * len(
                    self.reflections.select(d_spacings > d_min_indexed)
                )
                crystal_ids = self.reflections.select(d_spacings > d_min_indexed)["id"]
                if (crystal_ids == -1).count(True) < min_reflections_for_indexing:
                    logger.info(
                        "Finish searching for more lattices: %i unindexed reflections remaining.",
                        (crystal_ids == -1).count(True),
                    )
                    break

            n_lattices_previous_cycle = len(experiments)

            if self.d_min is None:
                self.d_min = self.params.refinement_protocol.d_min_start

            if len(experiments) == 0:
                new_expts = self.find_lattices()
                generate_experiment_identifiers(new_expts)
                experiments.extend(new_expts)
            else:
                try:
                    new = self.find_lattices()
                    generate_experiment_identifiers(new)
                    experiments.extend(new)
                except DialsIndexError:
                    logger.info("Indexing remaining reflections failed")

            if self.params.refinement_protocol.d_min_step is libtbx.Auto:
                n_cycles = self.params.refinement_protocol.n_macro_cycles
                if self.d_min is None or n_cycles == 1:
                    self.params.refinement_protocol.d_min_step = 0
                else:
                    d_spacings = 1 / self.reflections["rlp"].norms()
                    d_min_all = flex.min(d_spacings)
                    self.params.refinement_protocol.d_min_step = (
                        self.d_min - d_min_all
                    ) / (n_cycles - 1)
                    logger.info(
                        "Using d_min_step %.1f",
                        self.params.refinement_protocol.d_min_step,
                    )

            if len(experiments) == 0:
                raise DialsIndexError("No suitable lattice could be found.")
            elif len(experiments) == n_lattices_previous_cycle:
                logger.warning("No more suitable lattices could be found")
                # no more lattices found
                break

            for i_cycle in range(self.params.refinement_protocol.n_macro_cycles):
                if (
                    i_cycle > 0
                    and self.d_min is not None
                    and self.params.refinement_protocol.d_min_step > 0
                ):
                    d_min = self.d_min - self.params.refinement_protocol.d_min_step
                    d_min = max(d_min, 0)
                    if self.params.refinement_protocol.d_min_final is not None:
                        d_min = max(d_min, self.params.refinement_protocol.d_min_final)
                    if d_min >= 0:
                        self.d_min = d_min
                        logger.info("Increasing resolution to %.2f Angstrom", d_min)

                # reset reflection lattice flags
                # the lattice a given reflection belongs to: a value of -1 indicates
                # that a reflection doesn't belong to any lattice so far
                self.reflections["id"] = flex.int(len(self.reflections), -1)

                self.index_reflections(experiments, self.reflections)

                if i_cycle == 0 and self.params.known_symmetry.space_group is not None:
                    self._apply_symmetry_post_indexing(
                        experiments, self.reflections, n_lattices_previous_cycle
                    )

                logger.info("\nIndexed crystal models:")
                self.show_experiments(experiments, self.reflections, d_min=self.d_min)

                if self._check_have_similar_crystal_models(experiments):
                    have_similar_crystal_models = True
                    break

                logger.info("")
                logger.info("#" * 80)
                logger.info("Starting refinement (macro-cycle %i)", i_cycle + 1)
                logger.info("#" * 80)
                logger.info("")
                self.indexed_reflections = self.reflections["id"] > -1

                sel = flex.bool(len(self.reflections), False)
                lengths = 1 / self.reflections["rlp"].norms()
                if self.d_min is not None:
                    isel = (lengths <= self.d_min).iselection()
                    sel.set_selected(isel, True)
                sel.set_selected(self.reflections["id"] == -1, True)
                self.reflections.unset_flags(sel, self.reflections.flags.indexed)
                self.unindexed_reflections = self.reflections.select(sel)

                reflections_for_refinement = self.reflections.select(
                    self.indexed_reflections
                )
                if self.params.refinement_protocol.mode == "repredict_only":
                    refined_experiments, refined_reflections = (
                        experiments,
                        reflections_for_refinement,
                    )
                    from dials.algorithms.refinement.prediction.managed_predictors import (
                        ExperimentsPredictorFactory,
                    )

                    ref_predictor = ExperimentsPredictorFactory.from_experiments(
                        experiments,
                        spherical_relp=self.all_params.refinement.parameterisation.spherical_relp_model,
                    )
                    ref_predictor(refined_reflections)
                else:
                    try:
                        refined_experiments, refined_reflections = self.refine(
                            experiments, reflections_for_refinement
                        )
                    except (DialsRefineConfigError, DialsRefineRuntimeError) as e:
                        if len(experiments) == 1:
                            raise DialsIndexRefineError(str(e))
                        had_refinement_error = True
                        logger.info("Refinement failed:")
                        logger.info(e)
                        del experiments[-1]

                        # remove experiment id from the reflections associated
                        # with this deleted experiment - indexed flag removed
                        # below
                        last = len(experiments)
                        sel = refined_reflections["id"] == last
                        logger.info(
                            "Removing %d reflections with id %d", sel.count(True), last
                        )
                        refined_reflections["id"].set_selected(sel, -1)

                        break

                self._unit_cell_volume_sanity_check(experiments, refined_experiments)

                self.refined_reflections = refined_reflections
                self.refined_reflections.unset_flags(
                    self.refined_reflections["id"] < 0,
                    self.refined_reflections.flags.indexed,
                )

                for i, expt in enumerate(self.experiments):
                    ref_sel = self.refined_reflections.select(
                        self.refined_reflections["imageset_id"] == i
                    )
                    ref_sel = ref_sel.select(ref_sel["id"] >= 0)
                    for i_expt in set(ref_sel["id"]):
                        refined_expt = refined_experiments[i_expt]
                        expt.detector = refined_expt.detector
                        expt.beam = refined_expt.beam
                        expt.goniometer = refined_expt.goniometer
                        expt.scan = refined_expt.scan
                        refined_expt.imageset = expt.imageset

                if not (
                    self.all_params.refinement.parameterisation.beam.fix == "all"
                    and self.all_params.refinement.parameterisation.detector.fix
                    == "all"
                ):
                    # Experimental geometry may have changed - re-map centroids to
                    # reciprocal space
                    self.reflections.map_centroids_to_reciprocal_space(self.experiments)

                # update for next cycle
                experiments = refined_experiments
                self.refined_experiments = refined_experiments

                logger.info("\nRefined crystal models:")
                self.show_experiments(
                    self.refined_experiments, self.reflections, d_min=self.d_min
                )

                if (
                    i_cycle >= 2
                    and self.d_min == self.params.refinement_protocol.d_min_final
                ):
                    logger.info("Target d_min_final reached: finished with refinement")
                    break

        if self.refined_experiments is None:
            raise DialsIndexRefineError("None of the experiments could refine.")

        if len(self.refined_experiments) > 1:
            from dials.algorithms.indexing.compare_orientation_matrices import (
                rotation_matrix_differences,
            )

            logger.info(
                rotation_matrix_differences(self.refined_experiments.crystals())
            )

        self._xyzcal_mm_to_px(self.refined_experiments, self.refined_reflections)

    def _unit_cell_volume_sanity_check(self, original_experiments, refined_experiments):
        # sanity check for unrealistic unit cell volume increase during refinement
        # usually this indicates too many parameters are being refined given the
        # number of observations provided.
        if not self.params.refinement_protocol.disable_unit_cell_volume_sanity_check:
            for orig_expt, refined_expt in zip(
                original_experiments, refined_experiments
            ):
                uc1 = orig_expt.crystal.get_unit_cell()
                uc2 = refined_expt.crystal.get_unit_cell()
                volume_change = abs(uc1.volume() - uc2.volume()) / uc1.volume()
                cutoff = 0.5
                if volume_change > cutoff:
                    msg = "\n".join(
                        (
                            "Unrealistic unit cell volume increase during refinement of %.1f%%.",
                            "Please try refining fewer parameters, either by enforcing symmetry",
                            "constraints (space_group=) and/or disabling experimental geometry",
                            "refinement (detector.fix=all and beam.fix=all). To disable this",
                            "sanity check set disable_unit_cell_volume_sanity_check=True.",
                        )
                    ) % (100 * volume_change)
                    raise DialsIndexError(msg)

    def _apply_symmetry_post_indexing(
        self, experiments, reflections, n_lattices_previous_cycle
    ):
        # now apply the space group symmetry only after the first indexing
        # need to make sure that the symmetrized orientation is similar to the P1 model
        for cryst in experiments.crystals()[n_lattices_previous_cycle:]:
            new_cryst, cb_op = self._symmetry_handler.apply_symmetry(cryst)
            new_cryst = new_cryst.change_basis(cb_op)
            cryst.update(new_cryst)
            cryst.set_space_group(self.params.known_symmetry.space_group.group())
            for i_expt, expt in enumerate(experiments):
                if expt.crystal is not cryst:
                    continue
                if not cb_op.is_identity_op():
                    miller_indices = reflections["miller_index"].select(
                        reflections["id"] == i_expt
                    )
                    miller_indices = cb_op.apply(miller_indices)
                    reflections["miller_index"].set_selected(
                        reflections["id"] == i_expt, miller_indices
                    )

    def _check_have_similar_crystal_models(self, experiments):
        """
        Checks for similar crystal models.

        Checks whether the most recently added crystal model is similar to previously
        found crystal models, and if so, deletes the last crystal model from the
        experiment list.
        """
        have_similar_crystal_models = False
        cryst_b = experiments.crystals()[-1]
        for i_a, cryst_a in enumerate(experiments.crystals()[:-1]):
            R_ab, axis, angle, cb_op_ab = difference_rotation_matrix_axis_angle(
                cryst_a, cryst_b
            )
            min_angle = self.params.multiple_lattice_search.minimum_angular_separation
            if abs(angle) < min_angle:  # degrees
                logger.info(
                    "Crystal models too similar, rejecting crystal %i:",
                    len(experiments),
                )
                logger.info(
                    "Rotation matrix to transform crystal %i to crystal %i",
                    i_a + 1,
                    len(experiments),
                )
                logger.info(R_ab)
                logger.info(
                    f"Rotation of {angle:.3f} degrees"
                    + " about axis (%.3f, %.3f, %.3f)" % axis
                )
                have_similar_crystal_models = True
                del experiments[-1]
                break
        return have_similar_crystal_models

    def _xyzcal_mm_to_px(self, experiments, reflections):
        # set xyzcal.px field in reflections
        reflections["xyzcal.px"] = flex.vec3_double(len(reflections))
        for i, expt in enumerate(experiments):
            imgset_sel = reflections["imageset_id"] == i
            refined_reflections = reflections.select(imgset_sel)
            panel_numbers = flex.size_t(refined_reflections["panel"])
            xyzcal_mm = refined_reflections["xyzcal.mm"]
            x_mm, y_mm, z_rad = xyzcal_mm.parts()
            xy_cal_mm = flex.vec2_double(x_mm, y_mm)
            xy_cal_px = flex.vec2_double(len(xy_cal_mm))
            for i_panel in range(len(expt.detector)):
                panel = expt.detector[i_panel]
                sel = panel_numbers == i_panel
                xy_cal_px.set_selected(
                    sel, panel.millimeter_to_pixel(xy_cal_mm.select(sel))
                )
            x_px, y_px = xy_cal_px.parts()
            if expt.scan is not None:
                z_px = expt.scan.get_array_index_from_angle(z_rad, deg=False)
            else:
                # must be a still image, z centroid not meaningful
                z_px = z_rad
            xyzcal_px = flex.vec3_double(x_px, y_px, z_px)
            reflections["xyzcal.px"].set_selected(imgset_sel, xyzcal_px)

    def show_experiments(self, experiments, reflections, d_min=None):
        if d_min is not None:
            reciprocal_lattice_points = reflections["rlp"]
            d_spacings = 1 / reciprocal_lattice_points.norms()
            reflections = reflections.select(d_spacings > d_min)
        for i_expt, expt in enumerate(experiments):
            logger.info(
                "model %i (%i reflections):",
                i_expt + 1,
                (reflections["id"] == i_expt).count(True),
            )
            logger.info(expt.crystal)

        indexed_flags = reflections.get_flags(reflections.flags.indexed)
        imageset_id = reflections["imageset_id"]
        rows = [["Imageset", "# indexed", "# unindexed", "% indexed"]]
        for i in range(flex.max(imageset_id) + 1):
            imageset_indexed_flags = indexed_flags.select(imageset_id == i)
            indexed_count = imageset_indexed_flags.count(True)
            unindexed_count = imageset_indexed_flags.count(False)
            rows.append(
                [
                    str(i),
                    str(indexed_count),
                    str(unindexed_count),
                    f"{indexed_count / (indexed_count + unindexed_count):.1%}",
                ]
            )
        logger.info(dials.util.tabulate(rows, headers="firstrow"))

    def find_max_cell(self):
        params = self.params.max_cell_estimation
        if self.params.max_cell is libtbx.Auto:
            if self.params.known_symmetry.unit_cell is not None:
                uc_params = (
                    self._symmetry_handler.target_symmetry_primitive.unit_cell().parameters()
                )
                self.params.max_cell = params.multiplier * max(uc_params[:3])
                logger.info("Using max_cell: %.1f Angstrom", self.params.max_cell)
            else:
                self.params.max_cell = find_max_cell(
                    self.reflections,
                    max_cell_multiplier=params.multiplier,
                    step_size=params.step_size,
                    nearest_neighbor_percentile=params.nearest_neighbor_percentile,
                    histogram_binning=params.histogram_binning,
                    nn_per_bin=params.nn_per_bin,
                    max_height_fraction=params.max_height_fraction,
                    filter_ice=params.filter_ice,
                    filter_overlaps=params.filter_overlaps,
                    overlaps_border=params.overlaps_border,
                ).max_cell
                logger.info("Found max_cell: %.1f Angstrom", self.params.max_cell)

    def index_reflections(self, experiments, reflections):
        self._assign_indices(reflections, experiments, d_min=self.d_min)
        if self.hkl_offset is not None and self.hkl_offset != (0, 0, 0):
            reflections["miller_index"] = apply_hkl_offset(
                reflections["miller_index"], self.hkl_offset
            )
            self.hkl_offset = None

    def refine(self, experiments, reflections):
        from dials.algorithms.indexing.refinement import refine

        refiner, refined, outliers = refine(self.all_params, reflections, experiments)
        if outliers is not None:
            reflections["id"].set_selected(outliers, -1)
        predicted = refiner.predict_for_indexed()
        reflections["xyzcal.mm"] = predicted["xyzcal.mm"]
        reflections["entering"] = predicted["entering"]
        reflections.unset_flags(
            flex.bool(len(reflections), True), reflections.flags.centroid_outlier
        )
        assert (
            reflections.get_flags(reflections.flags.centroid_outlier).count(True) == 0
        )
        reflections.set_flags(
            predicted.get_flags(predicted.flags.centroid_outlier),
            reflections.flags.centroid_outlier,
        )
        reflections.set_flags(
            refiner.selection_used_for_refinement(),
            reflections.flags.used_in_refinement,
        )
        return refiner.get_experiments(), reflections

    def _debug_write_reciprocal_lattice_points_as_pdb(
        self, file_name="reciprocal_lattice.pdb"
    ):
        from cctbx import crystal, xray

        cs = crystal.symmetry(
            unit_cell=(1000, 1000, 1000, 90, 90, 90), space_group="P1"
        )
        for i_panel in range(len(self.experiments[0].detector)):
            if len(self.experiments[0].detector) > 1:
                file_name = "reciprocal_lattice_%i.pdb" % i_panel
            with open(file_name, "wb") as f:
                xs = xray.structure(crystal_symmetry=cs)
                reflections = self.reflections.select(
                    self.reflections["panel"] == i_panel
                )
                for site in reflections["rlp"]:
                    xs.add_scatterer(xray.scatterer("C", site=site))
                xs.sites_mod_short()
                f.write(xs.as_pdb_file())

    def export_as_json(
        self, experiments, file_name="indexed_experiments.json", compact=False
    ):
        assert experiments.is_consistent()
        experiments.as_file(file_name)

    def export_reflections(self, reflections, file_name="reflections.pickle"):
        reflections.as_file(file_name)

    def find_lattices(self):
        raise NotImplementedError()


def apply_hkl_offset(indices, offset):
    h, k, l = indices.as_vec3_double().parts()
    h += offset[0]
    k += offset[1]
    l += offset[2]
    return flex.miller_index(h.iround(), k.iround(), l.iround())
